from json import dumps as json_dumps

import pytest
from pydantic import BaseModel, Field, create_model

from mountaineer.client_builder.build_schemas import (
    OpenAPISchema,
    OpenAPIToTypescriptSchemaConverter,
)


class SubModel1(BaseModel):
    sub_a: str


class SubModel2(BaseModel):
    sub_b: int


class MyModel(BaseModel):
    a: str = Field(description="The a field")
    b: int
    c: SubModel1
    d: list[SubModel1]
    both_sub: list[SubModel1 | SubModel2 | None]
    sub_map: dict[str, SubModel1 | None]


def test_basic_interface():
    converter = OpenAPIToTypescriptSchemaConverter()
    result = converter.convert(MyModel)
    assert set(result.keys()) == {"MyModel", "SubModel1", "SubModel2", "Sub Map"}
    assert "interface MyModel {" in result["MyModel"]


def test_model_gathering():
    schema = OpenAPISchema(**MyModel.model_json_schema())

    converter = OpenAPIToTypescriptSchemaConverter()
    all_models = converter.gather_all_models(schema)

    # OpenAPI makes an object for the dictionary as well
    assert len(all_models) == 4
    assert {m.title for m in all_models} == {
        "SubModel1",
        "SubModel2",
        "MyModel",
        "Sub Map",
    }


@pytest.mark.parametrize(
    "python_type,expected_typescript_types",
    [
        (str, ["value: string"]),
        (int, ["value: number"]),
        (SubModel1, ["value: SubModel1"]),
        (list[SubModel1], ["value: Array<SubModel1>"]),
        (dict[str, SubModel1], ["value: Record<string, SubModel1>"]),
        (dict[str, int], ["value: Record<string, number>"]),
        ("SubModel1", ["value: SubModel1"]),
    ],
)
def test_python_to_typescript_types(
    python_type: type, expected_typescript_types: list[str]
):
    # Create a model schema automatically with the passed in typing
    fake_model = create_model(
        "FakeModel",
        value=(python_type, Field()),
    )

    schema = OpenAPISchema(**fake_model.model_json_schema())

    converter = OpenAPIToTypescriptSchemaConverter()
    interface_definition = converter.convert_schema_to_interface(schema, base=schema)

    for expected_str in expected_typescript_types:
        assert expected_str in interface_definition


def test_require_json_dictionaries():
    """
    JSON dictionaries can only serialize with string keys, so we need to make sure
    that the TypeScript interface enforces this.
    """

    class InvalidDictModel(BaseModel):
        value: dict[int, str]

    class ValidDictModel(BaseModel):
        value: dict[str, str]

    converter = OpenAPIToTypescriptSchemaConverter()
    converter.validate_typescript_candidate(ValidDictModel)

    with pytest.raises(ValueError):
        converter.validate_typescript_candidate(InvalidDictModel)


def test_get_model_json_schema_excludes_masked_fields():
    """
    Ensure we avoid serializing "excluded" fields.
    """

    class ExcludedModel(BaseModel):
        a: str

    class IncludedModel(BaseModel):
        b: str

    class MainModel(BaseModel):
        excluded_obj: ExcludedModel | None = Field(default=None, exclude=True)
        included_obj: IncludedModel | None

    builder = OpenAPIToTypescriptSchemaConverter()
    raw_openapi_result = json_dumps(MainModel.model_json_schema())
    fixed_openapi_result = json_dumps(builder.get_model_json_schema(MainModel))

    # First we check that pydantic actually does serialize excluded fields
    # If this starts failing, we should be able to remove our workaround
    assert "excluded_obj" in raw_openapi_result
    assert "ExcludedModel" in raw_openapi_result

    assert "excluded_obj" not in fixed_openapi_result
    assert "ExcludedModel" not in fixed_openapi_result
    assert "included_obj" in fixed_openapi_result
    assert "IncludedModel" in fixed_openapi_result



import pytest

from mountaineer.controller import ControllerBase
from mountaineer.render import (
    LinkAttribute,
    Metadata,
    RenderBase,
    ThemeColorMeta,
    ViewportMeta,
)


class StubRenderBase(RenderBase):
    pass


class StubController(ControllerBase):
    def render(self):
        return StubRenderBase()


@pytest.mark.parametrize(
    "metadata, expected_tags",
    [
        # Title
        (
            Metadata(
                title="MyTitle",
            ),
            ["<title>MyTitle</title>"],
        ),
        # Multiple links
        (
            Metadata(
                links=[
                    LinkAttribute(
                        rel="stylesheet",
                        href="/stylesheet1.css",
                    ),
                    LinkAttribute(
                        rel="stylesheet",
                        href="/stylesheet2.css",
                    ),
                ],
            ),
            [
                '<link rel="stylesheet" href="/stylesheet1.css" />',
                '<link rel="stylesheet" href="/stylesheet2.css" />',
            ],
        ),
        # Mixed meta tags
        (
            Metadata(
                metas=[
                    ThemeColorMeta(
                        color="#000000",
                    ),
                    ViewportMeta(
                        width="device-width",
                        initial_scale=1,
                        maximum_scale=2,
                        user_scalable=False,
                    ),
                ]
            ),
            [
                '<meta name="theme-color" content="#000000" />',
                '<meta name="viewport" content="width=device-width, initial-scale=1.0, maximum-scale=2.0, user-scalable=no" />',
            ],
        ),
    ],
)
def test_build_header(metadata: Metadata, expected_tags: list[str]):
    controller = StubController()
    assert controller.build_header(metadata) == expected_tags



from abc import ABC, abstractmethod
from inspect import getmembers, isawaitable, ismethod
from pathlib import Path
from re import compile as re_compile
from time import time
from typing import Any, Callable, Coroutine, Iterable

from fastapi.responses import HTMLResponse
from inflection import underscore

from mountaineer.actions import (
    FunctionActionType,
    FunctionMetadata,
    get_function_metadata,
)
from mountaineer.js_compiler.source_maps import SourceMapParser
from mountaineer.logging import LOGGER
from mountaineer.paths import ManagedViewPath
from mountaineer.render import Metadata, RenderBase, RenderNull
from mountaineer.ssr import V8RuntimeError, render_ssr


class ControllerBase(ABC):
    url: str
    # Typically, view paths should be a relative path to the local
    # Paths are only used if you need to specify an absolute path to another
    # file on disk
    view_path: str | ManagedViewPath

    bundled_scripts: list[str]

    def __init__(
        self, slow_ssr_threshold: float = 0.1, hard_ssr_timeout: float | None = 10.0
    ):
        """
        :param slow_ssr_threshold: Each python process has a single V8 runtime associated with
        it, so SSR rendering can become a bottleneck if it requires processing. We log a warning
        if we detect that an SSR render took longer than this threshold.
        :param hard_ssr_timeout: If the SSR render takes longer than this threshold, we will
        automatically kill the V8 runtime and return an error to the client. This is useful for
        avoiding blocking the reset of the server process if the React render logic hangs.

        """
        # Injected by the build framework
        self.bundled_scripts: list[str] = []
        self.ssr_path: Path | None = None
        self.initialized = True
        self.slow_ssr_threshold = slow_ssr_threshold
        self.hard_ssr_timeout = hard_ssr_timeout
        self.source_map: SourceMapParser | None = None

    @abstractmethod
    def render(
        self, *args, **kwargs
    ) -> RenderBase | None | Coroutine[Any, Any, RenderBase]:
        """
        Client implementations must override render() to define the data that will
        be pushed from the server to the client. This function must be typehinted with
        your response type:

        ```python
        class MyServerData(RenderBase):
            pass

        class MyController:
            def render(self) -> MyServerData:
                pass
        ```

        If you don't intend to sync any data from server->client you can typehint this function
        with an explicit None return annotation:

        ```python
        class MyController:
            def render(self) -> None:
                pass
        ```

        """
        pass

    async def _generate_html(self, *args, global_metadata: Metadata | None, **kwargs):
        if not self.ssr_path:
            # Try to resolve the path dynamically now
            raise ValueError("No SSR path set for this controller")

        # Because JSON is a subset of JavaScript, we can just dump the model as JSON and
        # insert it into the page.
        server_data = self.render(*args, **kwargs)
        if isawaitable(server_data):
            server_data = await server_data
        if server_data is None:
            server_data = RenderNull()

        # This isn't expected to happen, but we add a check to typeguard the following logic
        if not isinstance(server_data, RenderBase):
            raise ValueError(
                f"Controller.render() must return a RenderBase instance, not {type(server_data)}"
            )

        # If we got back metadata that includes a redirect, we should short-circuit the rest of the
        # render process and return a redirect response
        if server_data.metadata and server_data.metadata.explicit_response:
            return server_data.metadata.explicit_response

        metadatas: list[Metadata] = []
        if server_data.metadata:
            metadatas.append(server_data.metadata)
        if global_metadata and (
            server_data.metadata is None
            or not server_data.metadata.ignore_global_metadata
        ):
            metadatas.append(global_metadata)

        header_str = "\n".join(self.build_header(self.merge_metadatas(metadatas)))

        # Now that we've built the header, we can remove it from the server data
        # This makes our cache more efficient, since metadata changes don't affect
        # the actual page contents.
        server_data = server_data.model_copy(update={"metadata": None})

        # TODO: Provide a function to automatically sniff for the client view folder
        start = time()
        try:
            ssr_html = render_ssr(
                self.ssr_path.read_text(),
                server_data,
                hard_timeout=self.hard_ssr_timeout,
            )
        except V8RuntimeError as e:
            # Try to parse the file sources and re-raise the error with the
            # maps on the current filesystem
            if self.source_map is not None:
                # parse() is a no-op if the source map is already parsed, so we can do it again
                self.source_map.parse()
                raise V8RuntimeError(self.source_map.map_exception(str(e)))
            raise e

        ssr_duration = time() - start
        if ssr_duration > self.slow_ssr_threshold:
            LOGGER.warning(f"Slow SSR render detected: {ssr_duration:.2f}s")
        else:
            LOGGER.debug(f"SSR render took {ssr_duration:.2f}s")

        # Client-side react scripts that will hydrate the server side contents on load
        server_data_json = server_data.model_dump_json()
        optional_scripts = "\n".join(
            [
                f"<script src='/static/{script_name}'></script>"
                for script_name in self.bundled_scripts
            ]
        )

        page_contents = f"""
        <html>
        <head>
        {header_str}
        </head>
        <body>
        <div id="root">{ssr_html}</div>
        <script type="text/javascript">
        const SERVER_DATA = {server_data_json};
        </script>
        {optional_scripts}
        </body>
        </html>
        """

        return HTMLResponse(page_contents)

    def build_header(self, metadata: Metadata) -> list[str]:
        """
        Builds the header for this controller. Returns the list of tags that will be injected into the
        <head> tag of the rendered page.

        """
        tags: list[str] = []

        def format_optional_keys(payload: dict[str, str | None]) -> str:
            return " ".join(
                [
                    f'{key}="{value}"'
                    for key, value in payload.items()
                    if value is not None
                ]
            )

        if metadata.title:
            tags.append(f"<title>{metadata.title}</title>")

        for meta_definition in metadata.metas:
            all_attributes = {
                "name": meta_definition.name,
                "content": meta_definition.content,
                **meta_definition.optional_attributes,
            }
            tags.append(f"<meta {format_optional_keys(all_attributes)} />")

        for link_definition in metadata.links:
            all_attributes = {
                "rel": link_definition.rel,
                "href": link_definition.href,
                **link_definition.optional_attributes,
            }
            tags.append(f"<link {format_optional_keys(all_attributes)} />")

        return tags

    def _get_client_functions(self) -> Iterable[tuple[str, Callable, FunctionMetadata]]:
        """
        Returns all of the client-callable functions for this controller. Right now we force
        client accessible functions to either be wrapped by @sideeffect or @passthrough.

        """
        # Iterate over all the functions in this class and see which ones have a _metadata attribute
        for name, func in getmembers(self, predicate=ismethod):
            try:
                metadata = get_function_metadata(func)
                if metadata.action_type in {
                    FunctionActionType.PASSTHROUGH,
                    FunctionActionType.SIDEEFFECT,
                }:
                    yield name, func, metadata
            except AttributeError:
                continue

    def resolve_paths(self, view_base: Path):
        """
        Given the path to the root /view directory, resolve our various
        on-disk paths.

        """
        script_name = underscore(self.__class__.__name__)

        # The SSR path is going to be static
        ssr_path = view_base / "_ssr" / f"{script_name}.js"
        ssr_map_path = ssr_path.with_suffix(".js.map")
        self.ssr_path = ssr_path if ssr_path.exists() else None
        self.source_map = (
            SourceMapParser(ssr_map_path) if ssr_map_path.exists() else None
        )

        # Find the md5-converted cache path
        md5_script_pattern = re_compile(script_name + "-" + "[a-f0-9]{32}" + ".js")

        self.bundled_scripts = [
            path.name
            for path in (view_base / "_static").iterdir()
            if md5_script_pattern.match(path.name) and ".js.map" not in path.name
        ]
        LOGGER.debug(
            f"[{self.__class__.__name__}] Resolved paths... {self.bundled_scripts}"
        )

    def merge_metadatas(self, metadatas: list[Metadata]):
        """
        Merges a list of metadata objects, sorted by priority. Some fields will
        take the union (like scripts) - others will prioritize earlier entries (title).

        """
        base_metadata = Metadata()

        for metadata in metadatas:
            base_metadata.title = base_metadata.title or metadata.title

            base_metadata.metas.extend(metadata.metas)
            base_metadata.links.extend(metadata.links)

        return base_metadata



from mountaineer.app import AppController
from mountaineer.js_compiler.postcss import PostCSSBundler
from mountaineer.render import LinkAttribute, Metadata

from ci_webapp.controllers.complex import ComplexController
from ci_webapp.controllers.detail import DetailController
from ci_webapp.controllers.home import HomeController
from ci_webapp.views import get_view_path

controller = AppController(
    view_root=get_view_path(""),
    global_metadata=Metadata(
        links=[LinkAttribute(rel="stylesheet", href="/static/app_main.css")]
    ),
    custom_builders=[
        PostCSSBundler(),
    ],
)
controller.register(HomeController())
controller.register(DetailController())
controller.register(ComplexController())



from fastapi import Depends
from sqlalchemy.ext.asyncio import AsyncEngine, async_sessionmaker, create_async_engine

from mountaineer.database.config import DatabaseConfig
from mountaineer.dependencies import CoreDependencies


class DatabaseDependencies:
    """
    Dependencies for use in API endpoint routes.

    """

    @staticmethod
    def get_db(
        config: DatabaseConfig = Depends(
            CoreDependencies.get_config_with_type(DatabaseConfig)
        ),
    ):
        if not config.SQLALCHEMY_DATABASE_URI:
            raise RuntimeError("No SQLALCHEMY_DATABASE_URI set")

        return create_async_engine(str(config.SQLALCHEMY_DATABASE_URI))

    @staticmethod
    async def get_db_session(
        engine: AsyncEngine = Depends(get_db),
    ):
        session_maker = async_sessionmaker(engine, expire_on_commit=False)
        async with session_maker() as session:
            yield session



from re import finditer as re_finditer
from typing import cast

from pydantic import BaseModel

from mountaineer.cache import extended_lru_cache
from mountaineer.static import get_static_path


class V8RuntimeError(Exception):
    pass


def fix_exception_lines(*, exception: str, injected_script: str):
    """
    Since we create a synthetic script to run in the V8 runtime, the line numbers
    of the in-application stack trace will be offset by however long our
    injected script is. This function re-processed the stack trace to fix the
    line number.

    """
    offset_lines = injected_script.count("\n")

    text_replacements: dict[tuple[int, int], str] = {}
    for match in re_finditer(r"\(([<>A-Za-z0-9]+?):(\d+?):(\d+?)\)", exception):
        # Only replace the line numbers
        text_replacements[match.span(2)] = str(int(match.group(2)) - offset_lines)

    sorted_replacements = sorted(
        text_replacements.items(), key=lambda x: x[0][0], reverse=True
    )

    for (start, end), replacement in sorted_replacements:
        exception = exception[:start] + replacement + exception[end:]

    return exception


@extended_lru_cache(maxsize=128, max_size_mb=5)
def render_ssr(
    script: str, render_data: BaseModel, hard_timeout: int | float | None = None
) -> str:
    """
    Render the react component in the provided SSR javascript bundle. This file will
    be directly executed within the V8 runtime.

    To speed up requests for the same exact content in the same time (ie. same react and same data)
    we cache the result of the render_ssr_rust call.

    :raises TimeoutError: If the render takes longer than the hard_timeout

    """
    from mountaineer import mountaineer as mountaineer_rs  # type: ignore

    polyfill_script = get_static_path("ssr_polyfills.js").read_text()
    data_json = render_data.model_dump_json()

    injected_script = f"const SERVER_DATA = {data_json};\n{polyfill_script}\n"
    full_script = f"{injected_script}{script}"

    try:
        # Convert to milliseconds for the rust worker
        render_result = mountaineer_rs.render_ssr(
            full_script, int(hard_timeout * 1000) if hard_timeout else 0
        )
    except ConnectionAbortedError:
        raise TimeoutError("SSR render was interrupted after hard timeout")
    except ValueError as e:
        raise V8RuntimeError(
            fix_exception_lines(exception=str(e), injected_script=injected_script)
        )

    return cast(str, render_result)



import subprocess
from itertools import product
from pathlib import Path
from time import sleep

import pytest
from click import secho
from requests import get

from create_mountaineer_app.builder import (
    build_project,
    environment_from_metadata,
    should_copy_path,
)
from create_mountaineer_app.generation import ProjectMetadata
from create_mountaineer_app.io import get_free_port
from create_mountaineer_app.templates import get_template_path


@pytest.mark.parametrize(
    "input_path, expected_copy",
    [
        (
            Path("myproject/__pycache__/test.pyc"),
            False,
        ),
        (
            Path("myproject/.gitignore"),
            False,
        ),
        (
            Path("myproject/.git/HEAD"),
            False,
        ),
        (
            Path("myproject/regular_file.txt"),
            True,
        ),
    ],
)
def test_copy_path(input_path: Path, expected_copy: bool):
    assert should_copy_path(input_path) == expected_copy


@pytest.mark.parametrize(
    "use_poetry, use_tailwind",
    product(
        # Use poetry
        [False, True],
        # Use tailwind
        [False, True],
    ),
)
@pytest.mark.integration_tests
def test_valid_permutations(
    tmpdir: str,
    use_poetry: bool,
    use_tailwind: bool,
):
    """
    Ensures that regardless of the input parameters
    used to generate our new project, it will successfully
    run and return the expected endpoints.

    """
    # Assume the create-mountaineer-app project is in the same directory as the
    # main mountaineer package. We use a slight hack here assuming that our file
    # directory is oriented like "mountaineer/create_mountaineer_app/create_mountaineer_app/templates"
    # main package.
    main_mountaineer_path = get_template_path("../../../").resolve()

    # Verify this is pointing to where we expect
    if not (main_mountaineer_path / "pyproject.toml").exists():
        raise ValueError("Unable to find the main mountaineer package.")

    metadata = ProjectMetadata(
        project_name="my_project",
        author_name="John Appleseed",
        author_email="test@email.com",
        project_path=Path(tmpdir),
        use_poetry=use_poetry,
        use_tailwind=use_tailwind,
        mountaineer_dev_path=main_mountaineer_path,
    )

    build_project(metadata)

    app_test_port = get_free_port()
    secho(f"Found free port for test server: {app_test_port}", fg="green")

    # Launch docker to host the default database
    subprocess.run(
        ["docker-compose", "up", "-d"], cwd=metadata.project_path, check=True
    )

    environment = environment_from_metadata(metadata)

    # Make sure the required models are created
    create_db_process = environment.run_command(["createdb"], metadata.project_path)
    create_db_process.communicate()

    # Now launch the server in the background
    process = environment.run_command(
        ["runserver", "--port", str(app_test_port)], metadata.project_path
    )

    try:
        # Wait up to 10s for the server to start and be accessible
        max_wait = 10
        while max_wait > 0:
            try:
                response = get(f"http://localhost:{app_test_port}")
                if response.ok:
                    break
            except Exception:
                pass
            secho(f"Waiting for server to start (remaining: {max_wait})...")
            sleep(1)
            max_wait -= 1

        assert max_wait > 0, "Server start timed out."

        secho("Test server started successfully", fg="green")

        # Perform the fetch tests
        response = get(f"http://localhost:{app_test_port}")
        assert response.ok

        response = get(f"http://localhost:{app_test_port}/not_found")
        assert not response.ok
    finally:
        if process.returncode is None:
            process.terminate()
            process.wait()
            secho("Server shut down...")
        else:
            secho(f"Server exited with code {process.returncode}")

        secho("Shutting down docker...")
        subprocess.run(
            ["docker-compose", "down"], cwd=metadata.project_path, check=True
        )
        secho("Docker shut down successfully.")



from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from mountaineer.app import AppController
from mountaineer.client_builder.builder import ClientBuilder
from mountaineer.controller import ControllerBase


class ExampleHomeController(ControllerBase):
    url = "/"
    view_path = "/page.tsx"

    def render(self) -> None:
        return None


class ExampleDetailController(ControllerBase):
    url = "/detail/{detail_id}/"
    view_path = "/detail/page.tsx"

    def render(self) -> None:
        return None


@pytest.fixture(scope="function")
def simple_app_controller():
    with TemporaryDirectory() as temp_dir_name:
        temp_view_path = Path(temp_dir_name)
        (temp_view_path / "detail").mkdir()

        # Simple view files
        (temp_view_path / "page.tsx").write_text("")
        (temp_view_path / "detail" / "page.tsx").write_text("")

        app_controller = AppController(view_root=temp_view_path)
        app_controller.register(ExampleHomeController())
        app_controller.register(ExampleDetailController())
        yield app_controller


@pytest.fixture
def builder(simple_app_controller: AppController):
    return ClientBuilder(simple_app_controller)


def test_generate_static_files(builder: ClientBuilder):
    builder.generate_static_files()


def test_generate_model_definitions(builder: ClientBuilder):
    builder.generate_model_definitions()


def test_generate_action_definitions(builder: ClientBuilder):
    builder.generate_action_definitions()


def test_generate_view_definitions(builder: ClientBuilder):
    builder.generate_link_shortcuts()


def test_generate_link_aggregator(builder: ClientBuilder):
    builder.generate_link_aggregator()


def test_generate_view_servers(builder: ClientBuilder):
    builder.generate_view_servers()



from contextlib import AsyncExitStack, asynccontextmanager
from typing import Callable, Type, TypeVar

from fastapi import Request
from fastapi.dependencies.utils import get_dependant, solve_dependencies
from pydantic_settings import BaseSettings

from mountaineer.config import get_config

T = TypeVar("T", bound=BaseSettings)


class CoreDependencies:
    @staticmethod
    def get_config_with_type(required_type: Type[T]):
        def internal_dependency() -> T:
            config = get_config()
            if not isinstance(config, required_type):
                raise TypeError(
                    f"Expected config to inherit from {required_type}, {type(config)} is not a valid subclass"
                )
            return config

        return internal_dependency


@asynccontextmanager
async def get_function_dependencies(
    *,
    callable: Callable,
    url: str | None = None,
    request: Request | None = None,
):
    """
    Get the dependencies of a function. This will return the values that should
    be injected into the function. Provide as much metadata as possible, so we can
    resolve more accurate dependencies. If not provided, we will synthesize some values.

    """
    # Synthesize defaults
    if not url:
        url = "/synthetic"
    if not request:
        request = Request(
            scope={
                "type": "http",
                "path": url,
                "path_params": {},
                "query_string": "",
                "headers": [],
            }
        )

    # Synthetic request object as if we're coming from the original first page
    dependant = get_dependant(
        call=callable,
        path=url,
    )

    async with AsyncExitStack() as async_exit_stack:
        values, errors, background_tasks, sub_response, _ = await solve_dependencies(
            request=request,
            dependant=dependant,
            async_exit_stack=async_exit_stack,
        )
        if background_tasks:
            raise RuntimeError(
                "Background tasks are not supported when calling a static function, due to undesirable side-effects."
            )
        if errors:
            raise RuntimeError(
                f"Errors encountered while resolving dependencies: {errors}"
            )

        yield values



from pathlib import Path

import pytest
from pydantic.main import BaseModel

from mountaineer.actions.fields import FunctionActionType, get_function_metadata
from mountaineer.actions.passthrough import passthrough
from mountaineer.annotation_helpers import MountaineerUnsetValue
from mountaineer.app import AppController
from mountaineer.controller import ControllerBase
from mountaineer.render import RenderBase


def test_markup_passthrough():
    """
    Check that the @passthrough decorator extracts the expected
    data from our model definition.
    """

    class ExamplePassthroughModel(BaseModel):
        first_name: str

    class TestController(ControllerBase):
        @passthrough(response_model=ExamplePassthroughModel)
        def get_external_data(self):
            return dict(
                first_name="John",
            )

    metadata = get_function_metadata(TestController.get_external_data)
    assert metadata.action_type == FunctionActionType.PASSTHROUGH
    assert metadata.get_passthrough_model() == ExamplePassthroughModel
    assert metadata.function_name == "get_external_data"
    assert isinstance(metadata.reload_states, MountaineerUnsetValue)
    assert isinstance(metadata.render_model, MountaineerUnsetValue)
    assert isinstance(metadata.url, MountaineerUnsetValue)
    assert isinstance(metadata.return_model, MountaineerUnsetValue)
    assert isinstance(metadata.render_router, MountaineerUnsetValue)


class ExampleRenderModel(RenderBase):
    value_a: str
    value_b: str


@pytest.mark.asyncio
async def test_can_call_passthrough():
    class TestController(ControllerBase):
        url: str = "/test/{query_id}/"

        def __init__(self):
            super().__init__()
            self.counter = 0
            self.render_counts = 0

        def render(
            self,
            query_id: int,
        ) -> ExampleRenderModel:
            self.render_counts += 1
            return ExampleRenderModel(
                value_a="Hello",
                value_b="World",
            )

        @passthrough
        def call_passthrough(self, payload: dict):
            self.counter += 1
            return dict(status="success")

        @passthrough
        async def call_passthrough_async(self, payload: dict):
            self.counter += 1
            return dict(status="success")

    app = AppController(view_root=Path())
    controller = TestController()
    app.register(controller)

    return_value_sync = await controller.call_passthrough(
        {},
    )

    return_value_async = await controller.call_passthrough_async(
        {},
    )

    # The response payload should be the same both both sync and async endpoints
    expected_response = {
        "passthrough": {
            "status": "success",
        }
    }

    assert return_value_sync == expected_response
    assert return_value_async == expected_response

    assert controller.counter == 2

    # Our passthrough logic by definition should not re-render
    assert controller.render_counts == 0



from typing import Any

from pydantic import PostgresDsn, model_validator
from pydantic_settings import BaseSettings


class DatabaseConfig(BaseSettings):
    POSTGRES_HOST: str
    POSTGRES_USER: str
    POSTGRES_PASSWORD: str
    POSTGRES_DB: str
    POSTGRES_PORT: int = 5432
    SQLALCHEMY_DATABASE_URI: PostgresDsn | None = None

    @model_validator(mode="before")
    def build_db_connection(cls, values: Any) -> Any:
        if not values.get("SQLALCHEMY_DATABASE_URI"):
            values["SQLALCHEMY_DATABASE_URI"] = PostgresDsn.build(
                scheme="postgresql+asyncpg",
                username=values["POSTGRES_USER"],
                password=values["POSTGRES_PASSWORD"],
                host=values["POSTGRES_HOST"],
                port=int(values.get("POSTGRES_PORT", 5432)),
                path=values["POSTGRES_DB"],
            )
        return values



"""
Utilities for unit and integration tests
"""
from functools import wraps
from inspect import isawaitable, iscoroutinefunction, signature
from tempfile import NamedTemporaryFile
from time import time

import pytest

from mountaineer.logging import LOGGER


class ExecutionTooLong(Exception):
    pass


def benchmark_function(max_time_seconds: int | float):
    """
    Wrap test functions in a timer that will enforce that the core logic completes
    in less than `max_time_seconds` seconds. Injects `start_timing` and `end_timing` into
    the function kwargs that will be called. Client callers should call these to scope
    where they want us to measure the core logic.

    """
    import pyinstrument

    def wrapper_fn(test_func):
        # We want to remove our custom functions from the signature, since pytest will natively
        # try to inject fixtures in this place
        orig_sig = signature(test_func)
        new_params = [
            p
            for name, p in orig_sig.parameters.items()
            if name not in {"start_timing", "end_timing"}
        ]
        new_sig = orig_sig.replace(parameters=new_params)

        # Make sure that our function is a coroutine (if it is we assume user has decorated it with
        # pytest.mark.asyncio), otherwise we should raise an error because pytest won't know how
        # to run it
        if not iscoroutinefunction(test_func):
            raise Exception(
                f"Test function {test_func.__name__} is not a coroutine function. Please decorate it with pytest.mark.asyncio"
            )

        @wraps(test_func)
        async def wrapper(*args, **kwargs):
            bound = new_sig.bind(*args, **kwargs)
            bound.apply_defaults()

            # Try to run the test function regularly, and time it
            # Instrumenting profilers at this point typically will slow down execution time
            start: float | None = None
            end: float | None = None

            def start_timing():
                nonlocal start
                start = time()

            def end_timing():
                nonlocal end
                end = time()

            try:
                result = test_func(
                    *args, **kwargs, start_timing=start_timing, end_timing=end_timing
                )
                if isawaitable(result):
                    result = await result

                if start is None:
                    raise Exception("Test function did not call start_timing")
                if end is None:
                    raise Exception("Test function did not call end_timing")

                LOGGER.info(f"Test function took: {end - start}")

                if end - start > max_time_seconds:
                    raise ExecutionTooLong()

                return result

            except ExecutionTooLong as e:
                LOGGER.error(f"Test function failed due to: {e}")

                # This should already be true, but we want to be explicit to help mypy
                assert start is not None
                assert end is not None

                profiler = pyinstrument.Profiler()
                output_filename: str | None = None

                def start_timing():
                    nonlocal profiler
                    profiler.start()

                def end_timing():
                    nonlocal profiler
                    nonlocal output_filename
                    profiler.stop()
                    with NamedTemporaryFile(delete=False, suffix=".html") as file:
                        file.write(profiler.output_html().encode())
                    LOGGER.warning(f"PyInstrument profile saved at: {file.name}")
                    output_filename = file.name

                result = test_func(
                    *args, **kwargs, start_timing=start_timing, end_timing=end_timing
                )
                if isawaitable(result):
                    await result

                pytest.fail(
                    f"Test function failed in {end-start}s and profiles generated; Pyinstrument: {output_filename}"
                )

        wrapper.__signature__ = new_sig  # type: ignore
        return wrapper

    return wrapper_fn



from contextlib import asynccontextmanager
from pathlib import Path
from time import time
from unittest.mock import patch

import pytest
from fastapi import Depends, Request
from fastapi.testclient import TestClient
from pydantic.main import BaseModel
from starlette.datastructures import Headers

from mountaineer.__tests__.common import calculate_primes
from mountaineer.actions.fields import FunctionActionType, get_function_metadata
from mountaineer.actions.sideeffect import sideeffect
from mountaineer.annotation_helpers import MountaineerUnsetValue
from mountaineer.app import AppController
from mountaineer.controller import ControllerBase
from mountaineer.logging import LOGGER
from mountaineer.render import RenderBase


class ExampleRenderModel(RenderBase):
    value_a: str
    value_b: str


def test_markup_sideeffect():
    """
    Check that the @sideeffect decorator extracts the expected
    data from our model definition.
    """

    class ExamplePassthroughModel(BaseModel):
        first_name: str

    class TestController(ControllerBase):
        # We declare as complicated a payload as @sideeffect supports so we can
        # see the full amount of metadata properties that are set
        @sideeffect(
            response_model=ExamplePassthroughModel,
            reload=tuple([ExampleRenderModel.value_a]),
        )
        def sideeffect_and_return_data(self):
            return dict(
                first_name="John",
            )

    metadata = get_function_metadata(TestController.sideeffect_and_return_data)
    assert metadata.action_type == FunctionActionType.SIDEEFFECT
    assert metadata.get_passthrough_model() == ExamplePassthroughModel
    assert metadata.function_name == "sideeffect_and_return_data"
    assert metadata.reload_states == tuple([ExampleRenderModel.value_a])
    assert isinstance(metadata.render_model, MountaineerUnsetValue)
    assert isinstance(metadata.url, MountaineerUnsetValue)
    assert isinstance(metadata.return_model, MountaineerUnsetValue)
    assert isinstance(metadata.render_router, MountaineerUnsetValue)


class ControllerCommon(ControllerBase):
    url: str = "/test/{query_id}/"

    def __init__(self):
        super().__init__()
        self.counter = 0
        self.render_counts = 0

    @sideeffect
    def call_sideeffect(self, payload: dict):
        self.counter += 1

    @sideeffect
    async def call_sideeffect_async(self, payload: dict):
        self.counter += 1


async def call_sideeffect_common(controller: ControllerCommon):
    app = AppController(view_root=Path())
    app.register(controller)

    @asynccontextmanager
    async def mock_get_render_parameters(*args, **kwargs):
        yield {
            "query_id": 1,
        }

    # After our wrapper is called, our function is now async
    # Avoid the dependency resolution logic since that's tested separately
    with patch(
        "mountaineer.actions.sideeffect.get_render_parameters"
    ) as patched_get_render_params:
        patched_get_render_params.side_effect = mock_get_render_parameters

        # Even if the "request" is not required by our sideeffects, it's required
        # by the function injected by the sideeffect decorator.
        return_value_sync = await controller.call_sideeffect(
            {},
            request=Request({"type": "http"}),
        )

        return_value_async = await controller.call_sideeffect_async(
            {},
            request=Request({"type": "http"}),
        )

        # The response payload should be the same both both sync and async endpoints
        expected_response = {
            "sideeffect": ExampleRenderModel(
                value_a="Hello",
                value_b="World",
            ),
            "passthrough": None,
        }

        assert return_value_sync == expected_response
        assert return_value_async == expected_response

        assert controller.counter == 2
        assert controller.render_counts == 2


@pytest.mark.asyncio
async def test_can_call_sideeffect():
    """
    Ensure that we can call the sideeffect, which will in turn
    call the render function to get fresh data.
    """

    class ExampleController(ControllerCommon):
        def render(
            self,
            query_id: int,
        ) -> ExampleRenderModel:
            self.render_counts += 1
            return ExampleRenderModel(
                value_a="Hello",
                value_b="World",
            )

    await call_sideeffect_common(ExampleController())


@pytest.mark.asyncio
async def test_can_call_sideeffect_async_render():
    """
    Render functions can also work asynchronously.
    """

    class ExampleController(ControllerCommon):
        async def render(
            self,
            query_id: int,
        ) -> ExampleRenderModel:
            self.render_counts += 1
            return ExampleRenderModel(
                value_a="Hello",
                value_b="World",
            )

    await call_sideeffect_common(ExampleController())


@pytest.mark.asyncio
async def test_get_render_parameters():
    """
    Given a controller, reproduce the logic of FastAPI to sniff the render()
    function for dependencies and resolve them.

    """
    from mountaineer.actions.sideeffect import get_render_parameters

    found_cookie = None

    def grab_cookie_dependency(request: Request):
        nonlocal found_cookie
        found_cookie = request.cookies.get("test-cookie")
        return found_cookie

    class TestController(ControllerBase):
        url: str = "/test/{query_id}/"

        def render(
            self,
            query_id: int,
            cookie_dependency: str = Depends(grab_cookie_dependency),
        ) -> ExampleRenderModel:
            return ExampleRenderModel(
                value_a="Hello",
                value_b="World",
            )

    # We need to load this test controller to an actual application runtime
    # or else we don't have the render() metadata added
    app = AppController(view_root=Path())
    controller = TestController()
    app.register(controller)

    fake_request = Request(
        {
            "type": "http",
            "headers": Headers(
                {
                    "cookie": "test-cookie=cookie-value",
                    # Its important the referer aligns with the controller url, since that is expected
                    # to be the original view page that is calling this sub-function
                    "referer": "http://example.com/test/5/",
                }
            ).raw,
            "http_version": "1.1",
            "scheme": "",
            "client": "",
            "server": "",
            # The URL and method should both be different, to test whether we are able
            # to map the request to the correct endpoint
            "method": "POST",
            "url": "http://localhost/related_action_endpoint",
        }
    )

    async with get_render_parameters(controller, fake_request) as resolved_dependencies:
        assert resolved_dependencies == {
            "cookie_dependency": "cookie-value",
            "query_id": 5,
        }


@pytest.mark.parametrize(
    "use_experimental,min_time,max_time",
    [
        (False, 1, None),
        (True, None, 0.05),
    ],
)
def test_limit_codepath_experimental(
    use_experimental: bool,
    min_time: float | None,
    max_time: float | None,
):
    class ExampleController(ControllerBase):
        url: str = "/test/{query_id}/"

        def render(
            self,
            query_id: int,
        ) -> ExampleRenderModel:
            a = calculate_primes(10000)
            b = calculate_primes(1000000)
            return ExampleRenderModel(
                value_a=f"Hello {a}",
                value_b=f"World {b}",
            )

        @sideeffect(
            reload=(ExampleRenderModel.value_a,),
            experimental_render_reload=use_experimental,
        )
        def call_sideeffect(self, payload: dict):
            pass

    # We need to load this test controller to an actual application runtime
    # or else we don't have the render() metadata added
    app = AppController(view_root=Path())
    controller = ExampleController()
    app.register(controller)

    sideeffect_url = get_function_metadata(ExampleController.call_sideeffect).url
    assert isinstance(sideeffect_url, str)

    client = TestClient(app.app)
    start = time()
    response = client.post(
        sideeffect_url,
        json={},
        headers={
            # From the original view page that is calling this sub-function
            "referer": "http://example.com/test/5/",
        },
    )
    elapsed = time() - start
    assert response.status_code == 200
    assert response.json() == {
        "sideeffect": {
            "value_a": "Hello 1229",
        }
    }

    LOGGER.info(f"Use Experimental: {use_experimental}\nElapsed: {elapsed}")

    if min_time is not None:
        assert elapsed >= min_time
    if max_time is not None:
        assert elapsed <= max_time



from functools import wraps
from inspect import isclass, signature
from pathlib import Path
from typing import Any, Callable

from fastapi import APIRouter, FastAPI, HTTPException, Request
from fastapi.openapi.utils import get_openapi
from fastapi.responses import JSONResponse
from fastapi.staticfiles import StaticFiles
from inflection import underscore
from pydantic import BaseModel
from pydantic_settings import BaseSettings
from starlette.routing import BaseRoute

from mountaineer.actions import (
    FunctionActionType,
    fuse_metadata_to_response_typehint,
    init_function_metadata,
)
from mountaineer.annotation_helpers import MountaineerUnsetValue
from mountaineer.controller import ControllerBase
from mountaineer.exceptions import APIException, APIExceptionInternalModelBase
from mountaineer.js_compiler.base import ClientBuilderBase
from mountaineer.js_compiler.bundler import JavascriptBundler
from mountaineer.logging import LOGGER
from mountaineer.paths import ManagedViewPath
from mountaineer.render import Metadata


class ControllerDefinition(BaseModel):
    controller: ControllerBase
    router: APIRouter
    # URL prefix to the root of the server
    url_prefix: str
    # Dynamically generated function that actually renders the html content
    # This is a hybrid between render() and _generate_html()
    view_route: Callable

    model_config = {
        "arbitrary_types_allowed": True,
    }


class AppController:
    """
    Main entrypoint of a project web application.

    """

    builders: list[ClientBuilderBase]
    global_metadata: Metadata | None

    def __init__(
        self,
        *,
        name: str = "Mountaineer Webapp",
        version: str = "0.1.0",
        view_root: Path,
        global_metadata: Metadata | None = None,
        custom_builders: list[ClientBuilderBase] | None = None,
        config: BaseSettings | None = None,
    ):
        """
        :param global_metadata: Script and meta will be applied to every
            page rendered by this application. Title will only be applied
            if the page does not already have a title set.
        :param config: Application global configuration.

        """
        self.app = FastAPI(title=name, version=version)
        self.controllers: list[ControllerDefinition] = []
        self.name = name
        self.version = version
        self.view_root = ManagedViewPath.from_view_root(view_root)
        self.global_metadata = global_metadata
        self.builders = [
            # Default builders
            JavascriptBundler(),
            # Custom builders
            *(custom_builders if custom_builders else []),
        ]

        # The act of instantiating the config should register it with the
        # global settings registry. We keep a reference to it so we can shortcut
        # to the user-defined settings later, but this is largely optional.
        self.config = config

        self.internal_api_prefix = "/internal/api"

        # The static directory has to exist before we try to mount it
        static_dir = self.view_root.get_managed_static_dir()

        # Mount the view_root / _static directory, since we'll need
        # this for the client mounted view files
        self.app.mount(
            "/static",
            StaticFiles(directory=str(static_dir)),
            name="static",
        )

        self.app.exception_handler(APIException)(self.handle_exception)

        self.app.openapi = self.generate_openapi  # type: ignore

    def register(self, controller: ControllerBase):
        """
        Register a new controller. This will:
            - Mount the html of the controller to the main application service
            - Mount all actions (ie. @sideeffect and @passthrough decorated functions) to their public API

        """
        # Update the paths now that we have access to the runtime package path
        controller.resolve_paths(self.view_root)

        # The controller superclass needs to be initialized before it's
        # registered into the application
        if not hasattr(controller, "initialized"):
            raise ValueError(
                f"You must call super().__init__() on {controller} before it can be registered."
            )

        # We need to passthrough the API of the render function to the FastAPI router so it's called properly
        # with the dependency injection kwargs
        @wraps(controller.render)
        async def generate_controller_html(*args, **kwargs):
            try:
                return await controller._generate_html(
                    *args, global_metadata=self.global_metadata, **kwargs
                )
            except Exception as e:
                # If a user explicitly is raising an APIException, we don't want to log it
                if not isinstance(e, (APIExceptionInternalModelBase, HTTPException)):
                    # Forward along the exception, just modify it to include
                    # the controller name for additional context
                    LOGGER.error(
                        f"Exception encountered in {controller.__class__.__name__} rendering"
                    )
                raise

        # Strip the return annotations from the function, since we just intend to return an HTML page
        # and not a JSON response
        if not hasattr(generate_controller_html, "__wrapped__"):
            raise ValueError(
                "Unable to clear wrapped typehint, no wrapped function found."
            )

        return_model = generate_controller_html.__wrapped__.__annotations__.get(
            "return", MountaineerUnsetValue()
        )
        if isinstance(return_model, MountaineerUnsetValue):
            raise ValueError(
                "Controller render() function must have a return type annotation"
            )

        # Only the signature of the actual rendering function, not the original. We might
        # need to sniff render() again for its typehint
        generate_controller_html.__signature__ = signature(  # type: ignore
            generate_controller_html,
        ).replace(return_annotation=None)

        # Validate the return model is actually a RenderBase or explicitly marked up as None
        if not (
            return_model is None
            or (isclass(return_model) and issubclass(return_model, BaseModel))
        ):
            raise ValueError(
                "Controller render() return type annotation is not a RenderBase"
            )

        # Attach a new metadata wrapper to the original function so we can easily
        # recover it when attached to the class
        render_metadata = init_function_metadata(
            controller.render, FunctionActionType.RENDER
        )
        render_metadata.render_model = return_model

        # Register the rendering view to an isolated APIRoute, so we can keep track of its
        # the resulting router independently of the rest of the application
        # This is useful in cases where we need to do a render()->FastAPI lookup
        view_router = APIRouter()
        view_router.get(controller.url)(generate_controller_html)
        render_metadata.render_router = view_router
        self.app.include_router(view_router)

        # Create a wrapper router for each controller to hold the side-effects
        controller_api = APIRouter()
        controller_url_prefix = (
            f"{self.internal_api_prefix}/{underscore(controller.__class__.__name__)}"
        )
        for _, fn, metadata in controller._get_client_functions():
            # We need to delay adding the typehint for each function until we are here, adding the view. Since
            # decorators run before the class is actually mounted, they're isolated from the larger class/controller
            # context that the action function is being defined within. Here since we have a global view
            # of the controller (render function + actions) this becomes trivial
            metadata.return_model = fuse_metadata_to_response_typehint(
                metadata, render_metadata.get_render_model()
            )

            # Update the signature of the internal function, which fastapi will sniff for the return declaration
            # https://github.com/tiangolo/fastapi/blob/a235d93002b925b0d2d7aa650b7ab6d7bb4b24dd/fastapi/dependencies/utils.py#L207
            method_function: Callable = fn.__func__  # type: ignore
            method_function.__signature__ = signature(method_function).replace(  # type: ignore
                return_annotation=metadata.return_model
            )

            metadata.url = (
                f"{controller_url_prefix}/{metadata.function_name.strip('/')}"
            )

            controller_api.post(f"/{metadata.function_name}")(fn)

        # Originally we tried implementing a sub-router for the internal API that was registered in the __init__
        # But the application greedily copies all contents from the router when it's added via `include_router`, so this
        # resulted in our endpoints not being seen even after calls to `.register(). We therefore attach the new
        # controller router directly to the application, since this will trigger a new copy of the routes.
        self.app.include_router(
            controller_api,
            prefix=controller_url_prefix,
        )

        LOGGER.debug(f"Did register controller: {controller.__class__.__name__}")

        self.controllers.append(
            ControllerDefinition(
                controller=controller,
                router=controller_api,
                view_route=generate_controller_html,
                url_prefix=controller_url_prefix,
            )
        )

    async def handle_exception(self, request: Request, exc: APIException):
        return JSONResponse(
            status_code=exc.status_code,
            content=exc.internal_model.model_dump(),
        )

    def generate_openapi(self, routes: list[BaseRoute] | None = None):
        """
        Bundle custom user exceptions in the OpenAPI schema. By default
        endpoints just include the 422 Validation Error, but this allows
        for custom derived user methods.

        """
        openapi_base = get_openapi(
            title=self.name,
            version=self.version,
            routes=(routes if routes is not None else self.app.routes),
        )

        # Loop over the registered controllers and get the actions
        exceptions_by_url = {}
        for controller_definition in self.controllers:
            for (
                _,
                _,
                metadata,
            ) in controller_definition.controller._get_client_functions():
                exceptions_models = metadata.get_exception_models()
                if not exceptions_models:
                    continue
                exceptions_by_url[metadata.url] = [
                    (
                        model.status_code,
                        model.InternalModel.__name__,
                        model.InternalModel.model_json_schema(),
                    )
                    for model in exceptions_models
                ]

        # Add to the schemas dictionary first
        # TODO: Add validation that throws an error on duplicate names
        # TODO: Add validation that throws an error for duplicate error codes
        for url, exception_payloads in exceptions_by_url.items():
            # Not included in the specified routes, we should ignore this controller
            if url not in openapi_base["paths"]:
                continue

            for status_code, schema_name, schema in exception_payloads:
                other_definitions = {
                    definition_name: self._update_ref_path(definition)
                    for definition_name, definition in schema.pop("$defs", {}).items()
                }
                openapi_base["components"]["schemas"].update(other_definitions)
                openapi_base["components"]["schemas"][
                    schema_name
                ] = self._update_ref_path(schema)

                # All actions are "posts" by definition
                openapi_base["paths"][url]["post"]["responses"][str(status_code)] = {
                    "description": f"Custom Error: {schema_name}",
                    "content": {
                        "application/json": {
                            "schema": {"$ref": f"#/components/schemas/{schema_name}"}
                        }
                    },
                }

        return openapi_base

    def _update_ref_path(self, schema: Any):
        """
        The $ref values that come out of the model schema are tied to #/defs instead
        of the #/components/schemas. This function updates the schema to use the
        correct prefix for the final OpenAPI schema.

        """
        if isinstance(schema, dict):
            new_schema: dict[str, Any] = {}
            for key, value in schema.items():
                if key == "$ref":
                    schema_name = value.split("/")[-1]
                    new_schema[key] = f"#/components/schemas/{schema_name}"
                    continue
                elif key == "additionalProperties":
                    # If the value is "False", we need to remove the key
                    if value is False:
                        continue

                new_schema[key] = self._update_ref_path(value)
            return new_schema
        elif isinstance(schema, list):
            return [self._update_ref_path(value) for value in schema]
        else:
            return schema



from mountaineer import ConfigBase
from mountaineer.database import DatabaseConfig
from pydantic_settings import SettingsConfigDict

class AppConfig(ConfigBase, DatabaseConfig):

    model_config = SettingsConfigDict(env_file=(".env",))



from pathlib import Path

import pytest

from mountaineer.watch import ChangeEventHandler, PackageWatchdog


@pytest.mark.parametrize(
    "path, ignore_list, ignore_hidden, expected_ignore",
    [
        ("myproject/regular.py", ["__pycache__"], True, False),
        ("myproject/.hidden", [""], True, True),
        ("myproject/.hidden/subfile", [""], True, True),
        ("myproject/.hidden/subfile", [""], False, False),
        ("myproject/__cache__", ["__cache__"], True, True),
    ],
)
def test_ignore_path(
    path: str,
    ignore_list: list[str],
    ignore_hidden: bool,
    expected_ignore: bool,
):
    handler = ChangeEventHandler(
        [], ignore_list=ignore_list, ignore_hidden=ignore_hidden
    )
    assert handler.should_ignore_path(Path(path)) == expected_ignore


@pytest.mark.parametrize(
    "paths, expected_paths",
    [
        (
            # Unique paths
            ["myproject/regular.py", "myproject/.hidden"],
            ["myproject/regular.py", "myproject/.hidden"],
        ),
        (
            # Simple subdirectory
            ["myproject", "myproject/subdir"],
            ["myproject"],
        ),
        (
            # More specific directory first
            ["myproject/subdir1/subdir2", "myproject/subdir1"],
            ["myproject/subdir1"],
        ),
    ],
)
def test_merge_paths(paths: list[str], expected_paths: list[str]):
    handler = PackageWatchdog("mountaineer", [])
    assert set(handler.merge_paths(paths)) == {
        str(Path(path).absolute()) for path in expected_paths
    }



from pathlib import Path
from re import sub as re_sub

import pytest

from mountaineer import mountaineer as mountaineer_rs  # type: ignore
from mountaineer.__tests__.fixtures import get_fixture_path
from mountaineer.js_compiler.source_maps import (
    SourceMapParser,
    SourceMapSchema,
    get_cleaned_js_contents,
    make_source_map_paths_absolute,
    update_source_map_path,
)
from mountaineer.test_utilities import benchmark_function


@pytest.mark.parametrize(
    "input_str, expected_output",
    [
        # Single-line comments
        ("var x = 5; // This is a comment", "var x = 5;"),
        # Multi-line comments
        (
            "var y = 10; /* Multi\nLine\nComment */ var z = 15;",
            "var y = 10;  var z = 15;",
        ),
        # Mixed comments
        ("// Comment\nvar a = 1; /* Comment */", "var a = 1;"),
        # No comments
        ("var b = 2;", "var b = 2;"),
        # Comments with URLs
        ("// Visit http://example.com", ""),
        # Comments in strings
        (
            "var url = 'http://example.com'; // comment",
            "var url = 'http://example.com';",
        ),
        # Nested comments (non-standard but for testing)
        ("/* Comment /* Nested Comment */ End */", "End */"),
        # Empty string
        ("", ""),
        # Whitespace handling
        ("   // Comment\nvar c = 3;   ", "var c = 3;"),
    ],
)
def test_get_cleaned_js_contents(input_str: str, expected_output: str):
    assert get_cleaned_js_contents(input_str) == expected_output


@pytest.mark.parametrize(
    "input_str, replace_path, expected_output",
    [
        # Standard replacement
        (
            "var testing; //# sourceMappingURL=myfile.js.map",
            "final_path.js.map",
            "var testing; //# sourceMappingURL=final_path.js.map",
        ),
        # Replacement of independent map paths
        (
            "var testing; //# sourceMappingURL=first.js.map //# sourceMappingURL=second.js.map",
            "final_path.js.map",
            "var testing; //# sourceMappingURL=final_path.js.map //# sourceMappingURL=final_path.js.map",
        ),
    ],
)
def test_update_source_map_path(
    input_str: str, replace_path: str, expected_output: str
):
    assert update_source_map_path(input_str, replace_path) == expected_output


@pytest.mark.asyncio
@benchmark_function(0.2)
async def test_parse_source_map_parse(
    start_timing,
    end_timing,
):
    parser = SourceMapParser(get_fixture_path("home_controller_source_map.js.map"))
    start_timing()
    parser.parse()
    end_timing()
    assert parser.source_map

    source_filenames = {
        Path(source_path).name for source_path in parser.source_map.sources
    }
    assert source_filenames == {
        "react.development.js",
        "index.js",
        "scheduler.development.js",
        "react-dom.development.js",
        "client.js",
        "synthetic_client.tsx",
        "live_reload.ts",
        "page.tsx",
        "useServer.ts",
        "api.ts",
        "links.ts",
        "actions.ts",
    }
    assert parser.get_original_location(37, 13)


def test_map_exception():
    exception_stacktrace = """
    Example client error
        at Page (<anonymous>:12882:13)
        at renderWithHooks (<anonymous>:6333:26)

    """

    def make_map_metadata(
        line_number: int, column_number: int, source_line: int, source_column: int
    ) -> mountaineer_rs.MapMetadata:
        """
        Workaround to deal with rust's MapMetadata constructor
        only accepting line number and column number.

        """
        metadata = mountaineer_rs.MapMetadata(
            line_number=line_number,
            column_number=column_number,
        )
        metadata.source_index = 0
        metadata.source_line = source_line
        metadata.source_column = source_column
        metadata.symbol_index = 0
        return metadata

    parser = SourceMapParser("")
    parser.source_map = SourceMapSchema(
        version=3,
        sources=["test.ts"],
        names=["symbol_test1", "symbol_test2"],
        mappings="",
        sourceRoot="",
        file="test.js",
    )
    parser.parsed_mappings = {
        (12882, 13): make_map_metadata(
            line_number=12882,
            column_number=13,
            source_line=500,
            source_column=10,
        ),
        (6333, 26): make_map_metadata(
            line_number=6333,
            column_number=13,
            source_line=600,
            source_column=20,
        ),
    }

    assert re_sub(r"\s+", "", parser.map_exception(exception_stacktrace)) == re_sub(
        r"\s+",
        "",
        """
    Example client error
        at Page (test.ts:500:10)
        at renderWithHooks (test.ts:600:20)
    """,
    )


def test_make_source_map_paths_absolute():
    schema = SourceMapSchema(
        version=3,
        sources=["../../node_modules/sub_folder/test.ts"],
        names=["symbol_test1", "symbol_test2"],
        mappings="",
        sourceRoot="",
        file="test.js",
    )

    parsed_result = make_source_map_paths_absolute(
        schema.model_dump_json(),
        Path("/fakehome/user/project/app/fake/fake/test.js"),
    )

    assert SourceMapSchema.model_validate_json(parsed_result).sources == [
        "/fakehome/user/project/app/node_modules/sub_folder/test.ts"
    ]



from functools import wraps
from inspect import isawaitable
from typing import TYPE_CHECKING, Callable, Type, overload

from pydantic import BaseModel

from mountaineer.actions.fields import (
    FunctionActionType,
    handle_explicit_responses,
    init_function_metadata,
)
from mountaineer.exceptions import APIException

if TYPE_CHECKING:
    from mountaineer.controller import ControllerBase


@overload
def passthrough(
    *,
    response_model: Type[BaseModel] | None = None,
    exception_models: list[Type[APIException]] | None = None,
) -> Callable[[Callable], Callable]:
    ...


@overload
def passthrough(func: Callable) -> Callable:
    ...


def passthrough(*args, **kwargs):
    """
    By default, we mask out function return values to avoid leaking any unintended data to client applications. This
    decorator marks a function .

    :response_model: Like in FastAPI, the response model to use for this endpoint. If not provided, will
        try to convert the response object into the proper JSON response as-is.

    """

    def decorator_with_args(
        response_model: Type[BaseModel] | None,
        exception_models: list[Type[APIException]] | None,
    ):
        def wrapper(func: Callable):
            @wraps(func)
            async def inner(self: "ControllerBase", *func_args, **func_kwargs):
                response = func(self, *func_args, **func_kwargs)
                if isawaitable(response):
                    response = await response
                return handle_explicit_responses(dict(passthrough=response))

            metadata = init_function_metadata(inner, FunctionActionType.PASSTHROUGH)
            metadata.passthrough_model = response_model
            metadata.exception_models = exception_models
            return inner

        return wrapper

    if args and callable(args[0]):
        # It's used as @sideeffect without arguments
        func = args[0]
        return decorator_with_args(None, None)(func)
    else:
        # It's used as @passthrough(xyz=2) with arguments
        return decorator_with_args(
            response_model=kwargs.get("response_model"),
            exception_models=kwargs.get("exception_models"),
        )



from mountaineer.database.config import DatabaseConfig  # noqa: F401
from mountaineer.database.dependencies import DatabaseDependencies  # noqa: F401
from mountaineer.database.sqlmodel import Field, SQLModel  # noqa: F401
from mountaineer.database.validator import DatabaseValidator  # noqa: F401



from mountaineer.actions import passthrough, sideeffect  # noqa: F401
from mountaineer.app import AppController  # noqa: F401
from mountaineer.config import ConfigBase  # noqa: F401
from mountaineer.controller import ControllerBase  # noqa: F401
from mountaineer.dependencies import CoreDependencies  # noqa: F401
from mountaineer.exceptions import APIException  # noqa: F401
from mountaineer.js_compiler.postcss import PostCSSBundler  # noqa: F401
from mountaineer.paths import ManagedViewPath  # noqa: F401
from mountaineer.render import (
    LinkAttribute,  # noqa: F401
    MetaAttribute,  # noqa: F401
    Metadata,  # noqa: F401
    RenderBase,  # noqa: F401
    ThemeColorMeta,  # noqa: F401
    ViewportMeta,  # noqa: F401
)



import asyncio
from hashlib import md5
from pathlib import Path
from tempfile import TemporaryDirectory

from inflection import underscore
from pydantic import BaseModel

from mountaineer.controller import ControllerBase
from mountaineer.js_compiler.base import ClientBuilderBase, ClientBundleMetadata
from mountaineer.js_compiler.esbuild import ESBuildWrapper
from mountaineer.js_compiler.source_maps import (
    get_cleaned_js_contents,
    make_source_map_paths_absolute,
    update_source_map_path,
)
from mountaineer.logging import LOGGER
from mountaineer.paths import ManagedViewPath, generate_relative_import


class BundleOutput(BaseModel):
    client_entrypoint_path: Path
    client_compiled_contents: str
    client_source_map_contents: str
    server_entrypoint_path: Path
    server_compiled_contents: str
    server_source_map_contents: str


class JavascriptBundler(ClientBuilderBase):
    """
    Compile the client-written tsx/jsx to raw javascript files for execution as part
    of the SSR pipeline and client hydration.
    """

    def __init__(self, root_element: str = "root", environment: str = "development"):
        self.root_element = root_element
        self.environment = environment

    async def handle_file(
        self,
        file_path: ManagedViewPath,
        controller: ControllerBase | None,
        metadata: ClientBundleMetadata,
    ):
        if controller is None:
            # Require a controller for our bundling
            return None
        if file_path.suffix not in [".tsx", ".jsx"]:
            # We only know how to parse tsx and jsx files
            return None

        # We need to generate a relative import path from the view root to the current file
        root_path = file_path.get_package_root_link()
        static_dir = root_path.get_managed_static_dir()
        ssr_dir = root_path.get_managed_ssr_dir()
        bundle = await self.generate_js_bundle(
            current_path=file_path, metadata=metadata
        )

        # Write the compiled files to disk
        # Client-side scripts have to be provided a cache-invalidation suffix alongside
        # mapping the source map to the new script name
        controller_base = underscore(controller.__class__.__name__)
        content_hash = md5(
            get_cleaned_js_contents(bundle.client_compiled_contents).encode()
        ).hexdigest()
        script_name = f"{controller_base}-{content_hash}.js"
        map_name = f"{script_name}.map"

        # Map to the new script name
        contents = update_source_map_path(bundle.client_compiled_contents, map_name)

        (static_dir / script_name).write_text(contents)
        (static_dir / map_name).write_text(
            make_source_map_paths_absolute(
                bundle.client_source_map_contents, bundle.client_entrypoint_path
            )
        )

        ssr_path = ssr_dir / f"{controller_base}.js"
        ssr_path.write_text(bundle.server_compiled_contents)
        (ssr_path.with_suffix(".js.map")).write_text(
            make_source_map_paths_absolute(
                bundle.server_source_map_contents, bundle.server_entrypoint_path
            )
        )

    async def generate_js_bundle(
        self, current_path: ManagedViewPath, metadata: ClientBundleMetadata
    ) -> BundleOutput:
        # Before starting, make sure all the files are valid
        self.validate_page(
            page_path=current_path, view_root_path=current_path.get_root_link()
        )

        layout_paths = self.sniff_for_layouts(
            page_path=current_path, view_root_path=current_path.get_root_link()
        )

        # esbuild works on disk files
        with TemporaryDirectory() as temp_dir_name:
            temp_dir_path = Path(temp_dir_name)

            # Actually create the dist directory, since our relative path sniffing approach
            # prefers to work with directories that exist
            (temp_dir_path / "dist").mkdir()

            # The same endpoint definition is used for both SSR and the client build
            synthetic_payload = self.build_synthetic_endpoint(
                page_path=current_path,
                layout_paths=layout_paths,
                # This output path should be relative to the `synthetic_client` and `synthetic_server`
                # entrypoint files
                output_path=temp_dir_path,
            )

            client_entrypoint = self.build_synthetic_client_page(*synthetic_payload)
            ssr_entrypoint = self.build_synthetic_ssr_page(*synthetic_payload)

            client_entrypoint_path = temp_dir_path / "synthetic_client.tsx"
            server_entrypoint_path = temp_dir_path / "synthetic_server.tsx"

            self.link_project_files(
                view_root_path=current_path.get_package_root_link(),
                temp_dir_path=temp_dir_path,
            )
            client_entrypoint_path.write_text(client_entrypoint)
            server_entrypoint_path.write_text(ssr_entrypoint)

            common_loader = {
                ".tsx": "tsx",
                ".jsx": "jsx",
            }

            es_builder = ESBuildWrapper()
            await asyncio.gather(
                *[
                    es_builder.bundle(
                        entry_points=[client_entrypoint_path],
                        outfile=temp_dir_path / "dist" / "synthetic_client.js",
                        output_format="esm",
                        bundle=True,
                        sourcemap=True,
                        define={
                            "process.env.NODE_ENV": self.environment,
                            "process.env.SSR_RENDERING": "false",
                            **(
                                {
                                    "process.env.LIVE_RELOAD_PORT": str(
                                        metadata.live_reload_port
                                    ),
                                }
                                if metadata.live_reload_port
                                else {}
                            ),
                        },
                        loaders=common_loader,
                        node_paths=[temp_dir_path / "node_modules"],
                    ),
                    es_builder.bundle(
                        entry_points=[server_entrypoint_path],
                        outfile=temp_dir_path / "dist" / "synthetic_server.js",
                        output_format="iife",
                        global_name="SSR",
                        define={
                            "global": "window",
                            "process.env.SSR_RENDERING": "true",
                            "process.env.NODE_ENV": self.environment,
                        },
                        bundle=True,
                        sourcemap=True,
                        loaders=common_loader,
                        node_paths=[temp_dir_path / "node_modules"],
                    ),
                ]
            )

            # Read these files
            return BundleOutput(
                client_entrypoint_path=client_entrypoint_path,
                client_compiled_contents=(
                    temp_dir_path / "dist" / "synthetic_client.js"
                ).read_text(),
                client_source_map_contents=(
                    temp_dir_path / "dist" / "synthetic_client.js.map"
                ).read_text(),
                server_entrypoint_path=server_entrypoint_path,
                server_compiled_contents=(
                    temp_dir_path / "dist" / "synthetic_server.js"
                ).read_text(),
                server_source_map_contents=(
                    temp_dir_path / "dist" / "synthetic_server.js.map"
                ).read_text(),
            )

    def build_synthetic_client_page(
        self,
        synthetic_imports: list[str],
        synthetic_endpoint: str,
        synthetic_endpoint_name: str,
    ):
        lines: list[str] = []

        # Assume the client page is always being called from a page that has been
        # initially rendered by SSR
        lines.append("import * as React from 'react';")
        lines.append("import { hydrateRoot } from 'react-dom/client';")
        lines += synthetic_imports

        lines.append(synthetic_endpoint)

        lines.append(
            f"const container = document.getElementById('{self.root_element}');"
        )
        lines.append(f"hydrateRoot(container, <{synthetic_endpoint_name} />);")

        return "\n".join(lines)

    def build_synthetic_ssr_page(
        self,
        synthetic_imports: list[str],
        synthetic_endpoint: str,
        synthetic_endpoint_name: str,
    ):
        lines: list[str] = []

        lines.append("import * as React from 'react';")
        lines.append("import { renderToString } from 'react-dom/server';")
        lines += synthetic_imports

        lines.append(synthetic_endpoint)

        lines.append(
            f"export const Index = () => renderToString(<{synthetic_endpoint_name} />);"
        )

        return "\n".join(lines)

    def build_synthetic_endpoint(
        self, *, page_path: ManagedViewPath, layout_paths: list[Path], output_path: Path
    ):
        """
        Following the Next.js syntax, layouts wrap individual pages in a top-down order. Here we
        create a synthetic page that wraps the actual page in the correct order.
        The output is a valid React file that acts as the page entrypoint
        for the `rootElement` ID in the DOM.

        """
        # All import paths have to be a relative path from the scratch directory
        # to the original file
        import_paths: list[str] = []

        static_api_path = (
            page_path.get_package_root_link().get_managed_code_dir() / "live_reload.ts"
        )
        import_paths.append(
            f"import mountLiveReload from '{generate_relative_import(output_path, static_api_path)}';"
        )

        import_paths.append(
            f"import Page from '{generate_relative_import(output_path, page_path)}';"
        )

        for i, layout_path in enumerate(layout_paths):
            import_paths.append(
                f"import Layout{i} from '{generate_relative_import(output_path, layout_path)}';"
            )

        # The synthetic endpoint is a function that returns a React component
        entrypoint_name = "Entrypoint"
        content_lines = [
            f"const {entrypoint_name} = () => {{",
            # This hook will only run in the browser, in dev mode
            # Otherwise it will short-circuit so it won't apply to production
            "mountLiveReload({});" "return (",
            *[f"<Layout{i}>" for i in range(len(layout_paths))],
            "<Page />",
            *[f"</Layout{i}>" for i in reversed(range(len(layout_paths)))],
            ");",
            "};",
        ]

        return import_paths, "\n".join(content_lines), entrypoint_name

    def link_project_files(self, *, view_root_path: Path, temp_dir_path: Path):
        """
        Javascript packages define a variety of build metadata in the root directory
        of the project (tsconfig.json, package.json, etc). Since we're running our esbuild pipeline
        in a temporary directory, we need to copy over the key files. We use a symbolic link
        to avoid copying the files over.
        """
        required_to_link = ["package.json", "node_modules"]
        optional_to_link = ["tsconfig.json"]

        for file_name in required_to_link + optional_to_link:
            # Only throw an error if the file is required to exist
            if not (view_root_path / file_name).exists():
                if file_name in required_to_link:
                    raise ValueError(
                        f"Error linking, expected {file_name} to exist in {view_root_path}"
                    )
                continue

            LOGGER.debug(f"Linking {file_name} to {temp_dir_path}")
            (temp_dir_path / file_name).symlink_to(view_root_path / file_name)

    def sniff_for_layouts(self, *, page_path: Path, view_root_path: Path):
        """
        Given a page.tsx path, find all the layouts that apply to it.
        Returns the layout paths that are found. Orders them from the top->down
        as they expect to be rendered.

        """
        # It's easier to handle absolute paths when doing direct string comparisons
        # of the file hierarchy.
        page_path = page_path.resolve().absolute()
        view_root_path = view_root_path.resolve().absolute()

        # Starting at the page path, walk up the directory tree and yield each layout
        # that is found.
        layouts: list[Path] = []
        current_path = page_path.parent

        while current_path != view_root_path:
            layout_path_tsx = current_path / "layout.tsx"
            layout_path_jsx = current_path / "layout.jsx"

            # We shouldn't have both in the same directory
            if layout_path_tsx.exists() and layout_path_jsx.exists():
                raise ValueError(
                    f"Duplicate layout definitions: {layout_path_tsx}, {layout_path_jsx}"
                )

            if layout_path_tsx.exists():
                layouts.append(layout_path_tsx)
            if layout_path_jsx.exists():
                layouts.append(layout_path_jsx)

            current_path = current_path.parent

        # Return the layouts in the order they should be rendered
        return list(reversed(layouts))

    def validate_page(self, *, page_path: Path, view_root_path: Path):
        # Validate that we're actually calling on a path file
        if page_path.name not in {"page.tsx", "page.jsx"}:
            raise ValueError(
                f"Invalid page path; view need to be specified in a `page.tsx` file: {page_path}"
            )

        # Validate that the page_path is within the view root. The following
        # logic assumes a hierarchical relationship between the two.
        if not page_path.is_relative_to(view_root_path):
            raise ValueError(
                f"Invalid page path, not relative to view root: {page_path} (root: {view_root_path})"
            )



from types import ModuleType
from typing import overload

from fastapi import Depends
from sqlalchemy.ext.asyncio import AsyncEngine

from mountaineer.database.dependencies import DatabaseDependencies
from mountaineer.database.sqlmodel import SQLModel
from mountaineer.dependencies import get_function_dependencies


@overload
async def handle_createdb(model_module: ModuleType) -> None:
    ...


@overload
async def handle_createdb(models: list[SQLModel]) -> None:
    ...


async def handle_createdb(*args, **kwargs):
    """
    Strictly speaking, passing a list of models isn't required. We just encourage
    an explicit passing of either the models module or the SQLModels themselves to make
    sure they are in-scope of the table registry when this function is run.

    """

    async def run_migrations(
        engine: AsyncEngine = Depends(DatabaseDependencies.get_db),
    ):
        async with engine.begin() as connection:
            await connection.run_sync(SQLModel.metadata.create_all)

    async with get_function_dependencies(callable=run_migrations) as values:
        await run_migrations(**values)



from dataclasses import dataclass
from pathlib import Path
from re import finditer as re_finditer
from re import sub
from time import time

from pydantic import BaseModel

from mountaineer.logging import LOGGER


@dataclass
class ValueMask:
    mask: int
    right_padding: int


class SourceMapSchema(BaseModel):
    version: int
    sources: list[str]
    names: list[str]
    mappings: str
    sourcesContent: list[str] | None = None
    sourceRoot: str | None = None
    file: str | None = None


class SourceMapParser:
    """
    Parse sourcemaps according to the official specification:
    https://sourcemaps.info/spec.html

    """

    def __init__(self, path: str | Path):
        """
        :param relative_to: If specified, will output source paths relative
        to this path.

        """
        from mountaineer import mountaineer as mountaineer_rs  # type: ignore

        self.path = Path(path)

        self.source_map: SourceMapSchema | None = None

        # { (line, column) : MapMetadata }
        self.parsed_mappings: dict[
            tuple[int, int], mountaineer_rs.MapMetadata
        ] | None = None

    def parse(self):
        from mountaineer import mountaineer as mountaineer_rs  # type: ignore

        # If we've already parsed this file, don't do it again
        if self.parsed_mappings is not None:
            return

        start_parse = time()
        self.source_map = SourceMapSchema.model_validate_json(
            Path(self.path).read_text()
        )
        LOGGER.debug(f"Parsed source map in {time() - start_parse:.2f}s")

        start_parse = time()
        self.parsed_mappings = mountaineer_rs.parse_source_map_mappings(
            self.source_map.mappings
        )
        LOGGER.debug(f"Parsed mappings in {time() - start_parse:.2f}s")

    def get_original_location(self, line: int, column: int):
        if self.parsed_mappings is None:
            raise ValueError("SourceMapParser has not been parsed yet")

        return self.parsed_mappings.get((line, column))

    def map_exception(self, exception: str):
        """
        Given a JS stack exception, try to map it to the original files and line numbers
        """
        if self.source_map is None or self.parsed_mappings is None:
            raise ValueError("SourceMapParser has not been parsed yet")

        # Build up the replacements all at once, since the matched indexes will be tied
        # to the original string
        text_replacements: dict[tuple[int, int], str] = {}
        for match in re_finditer(r"\(([<>A-Za-z0-9]+?):(\d+?):(\d+?)\)", exception):
            original_match = self.get_original_location(
                int(match.group(2)), int(match.group(3))
            )

            if original_match and original_match.source_index is not None:
                text_replacements[match.span(1)] = self.convert_relative_path(
                    self.source_map.sources[original_match.source_index]
                )
                text_replacements[match.span(2)] = str(original_match.source_line)
                text_replacements[match.span(3)] = str(original_match.source_column)

        # Sort in reverse order based on their start index to ensure that modifying parts of the string
        # doesn't affect the positions of parts that haven't been modified yet
        sorted_replacements = sorted(
            text_replacements.items(), key=lambda x: x[0][0], reverse=True
        )

        for (start, end), replacement in sorted_replacements:
            exception = exception[:start] + replacement + exception[end:]

        return exception

    def convert_relative_path(self, absolute_path: str):
        """
        If we're within a parent directory of path, make it relative to the current
        working directory. Otherwise just default to the original absolute path.

        """
        source_path = Path(absolute_path)

        if source_path.is_relative_to(Path.cwd()):
            return "./" + str(source_path.relative_to(Path.cwd()))

        return absolute_path


def get_cleaned_js_contents(contents: str):
    """
    Strip all single or multiline comments, since these can be dynamically generated
    metadata and can change without the underlying logic changing.
    """
    # Regular expression to match single line and multiline comments
    # This regex handles single-line comments (// ...), multi-line comments (/* ... */),
    # and avoids capturing URLs like http://...
    # It also considers edge cases where comment-like patterns are inside strings
    pattern = r"(\/\*[\s\S]*?\*\/|([^:]|^)\/\/[^\r\n]*)"

    # Using re.sub to replace the matched comments with an empty string
    return sub(pattern, "", contents).strip()


def update_source_map_path(contents: str, new_path: str):
    """
    Updates the source map path to the new path, since the path is dynamic.
    """
    return sub(r"sourceMappingURL=(.*?).map", f"sourceMappingURL={new_path}", contents)


def make_source_map_paths_absolute(contents: str, original_script_path: Path):
    """
    Takes a source map, along with the original pre-compiled entrypoint path,
    and transforms the relative paths sources into absolute paths. Since often
    our precompiled endpoints are in tmp directories, this is helpful to encode the
    persistent path to the source files.

    """
    payload = SourceMapSchema.model_validate_json(contents)

    new_sources: list[str] = []
    for source in payload.sources:
        source_path = Path(source)
        if not source_path.is_absolute():
            source_path = original_script_path.parent / source_path
            new_sources.append(str(source_path.resolve()))

    payload.sources = new_sources
    return payload.model_dump_json()



from .fields import (
    FunctionActionType,  # noqa
    FunctionMetadata,  # noqa
    fuse_metadata_to_response_typehint,  # noqa
    get_function_metadata,  # noqa
    init_function_metadata,  # noqa
)
from .passthrough import passthrough  # noqa
from .sideeffect import sideeffect  # noqa



import importlib.metadata
import os
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Flag, auto
from json import loads as json_loads
from pathlib import Path
from re import search as re_search
from threading import Timer
from typing import Callable

from click import secho
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer


class CallbackType(Flag):
    CREATED = auto()
    MODIFIED = auto()
    DELETED = auto()


@dataclass
class CallbackDefinition:
    action: CallbackType
    callback: Callable


class ChangeEventHandler(FileSystemEventHandler):
    def __init__(
        self,
        callbacks: list[CallbackDefinition],
        ignore_list=["__pycache__", "_ssr", "_static", "_server"],
        ignore_hidden=True,
        debounce_interval=0.1,
    ):
        """
        :param debounce_interval: Seconds to wait for more events. Will only send one event per batched
        interval to avoid saturating clients with one action that results in many files.

        """
        super().__init__()
        self.callbacks = callbacks
        self.ignore_changes = False
        self.ignore_list = ignore_list
        self.ignore_hidden = ignore_hidden
        self.debounce_interval = debounce_interval
        self.debounce_timer: Timer | None = None

    def on_modified(self, event):
        super().on_modified(event)
        if self.should_ignore_path(event.src_path):
            return
        if not event.is_directory:
            secho(f"File modified: {event.src_path}", fg="yellow")
            self._debounce(CallbackType.MODIFIED)

    def on_created(self, event):
        super().on_created(event)
        if self.should_ignore_path(event.src_path):
            return
        if not event.is_directory:
            secho(f"File created: {event.src_path}", fg="yellow")
            self._debounce(CallbackType.CREATED)

    def on_deleted(self, event):
        super().on_deleted(event)
        if self.should_ignore_path(event.src_path):
            return
        if not event.is_directory:
            secho(f"File deleted: {event.src_path}", fg="yellow")
            self._debounce(CallbackType.DELETED)

    def _debounce(self, action: CallbackType):
        if self.debounce_timer is not None:
            self.debounce_timer.cancel()
        self.debounce_timer = Timer(
            self.debounce_interval, self.handle_callbacks, [action]
        )
        self.debounce_timer.start()

    def handle_callbacks(self, action: CallbackType):
        """
        Runs all callbacks for the given action. Since callbacks are allowed to make
        modifications to the filesystem, we temporarily disable the event handler to avoid
        infinite loops.

        """
        self.ignore_changes = True
        for callback in self.callbacks:
            if action in callback.action:
                callback.callback()
        self.ignore_changes = False

    def should_ignore_path(self, path: Path):
        if self.ignore_changes:
            return True

        path_components = set(str(path).split("/"))

        # Check for any nested hidden directories and ignored directories
        if self.ignore_hidden and any(
            (component for component in path_components if component.startswith("."))
        ):
            return True
        elif path_components & set(self.ignore_list) != set():
            return True
        return False


class WatchdogLockError(Exception):
    def __init__(self, lock_path: Path):
        super().__init__(
            f"Watch lock file exists, another process may be running. If you're sure this is not the case, run:\n"
            f"`$ rm {lock_path}`",
        )
        self.lock_path = lock_path


class PackageWatchdog:
    def __init__(
        self,
        main_package: str,
        dependent_packages: list[str],
        callbacks: list[CallbackDefinition] | None = None,
        run_on_bootup: bool = False,
    ):
        """
        :param run_on_bootup: Typically, we will only notify callback if there has been
            a change to the filesystem. If this is set to True, we will run all callbacks
            on bootup as well.
        """
        self.main_package = main_package
        self.packages = [main_package] + dependent_packages
        self.paths: list[str] = []
        self.callbacks: list[CallbackDefinition] = callbacks or []
        self.run_on_bootup = run_on_bootup

        self.check_packages_installed()
        self.get_package_paths()

    def start_watching(self):
        with self.acquire_watchdog_lock():
            if self.run_on_bootup:
                for callback_definition in self.callbacks:
                    callback_definition.callback()

            event_handler = ChangeEventHandler(callbacks=self.callbacks)
            observer = Observer()

            for path in self.paths:
                secho(f"Watching {path}", fg="green")
                if os.path.isdir(path):
                    observer.schedule(event_handler, path, recursive=True)
                else:
                    observer.schedule(
                        event_handler, os.path.dirname(path), recursive=False
                    )

            observer.start()

            try:
                secho("Starting observer...")
                observer.join()
            except KeyboardInterrupt:
                observer.stop()
                observer.join()

    def check_packages_installed(self):
        for package in self.packages:
            try:
                importlib.metadata.version(package)
            except importlib.metadata.PackageNotFoundError:
                raise ValueError(
                    f"Package '{package}' is not installed in the current environment"
                )

    def get_package_paths(self):
        paths: list[str] = []
        for package in self.packages:
            dist = importlib.metadata.distribution(package)
            package_path = self.resolve_package_path(dist)
            paths.append(str(package_path))
        self.paths = self.merge_paths(paths)

    def resolve_package_path(self, dist: importlib.metadata.Distribution):
        """
        Given a package distribution, returns the local file directory that should be watched
        """
        # Recent versions of poetry install development packages (-e .) as direct URLs
        # https://the-hitchhikers-guide-to-packaging.readthedocs.io/en/latest/introduction.html
        # "Path configuration files have an extension of .pth, and each line must
        # contain a single path that will be appended to sys.path."
        package_name = dist.name.replace("-", "_")
        symbolic_links = [
            path for path in (dist.files or []) if path.name == f"{package_name}.pth"
        ]
        dist_links = [
            path
            for path in (dist.files or [])
            if path.name == "direct_url.json"
            and re_search(package_name + r"-[0-9-.]+\.dist-info", path.parent.name)
        ]
        explicit_links = [
            path
            for path in (dist.files or [])
            if path.parent.name == package_name
            and (
                # Sanity check that the parent is the high level project directory
                # by looking for common base files
                path.name == "__init__.py"
            )
        ]

        if symbolic_links:
            direct_url_path = symbolic_links[0]
            return dist.locate_file(direct_url_path.read_text().strip())

        if dist_links:
            dist_link = dist_links[0]
            direct_metadata = json_loads(dist_link.read_text())
            package_path = "/" + direct_metadata["url"].lstrip("file://").lstrip("/")
            return dist.locate_file(package_path)

        if explicit_links:
            # Since we found the __init__.py file for the root, we should be able to go up
            # to the main path
            explicit_link = explicit_links[0]
            return dist.locate_file(explicit_link.parent)

        raise ValueError(
            f"Could not find a valid path for package {dist.name}, found files: {dist.files}"
        )

    @contextmanager
    def acquire_watchdog_lock(self):
        """
        We only want one watchdog running at a time or otherwise we risk stepping
        on each other or having infinitely looping file change notifications.

        """
        dist = importlib.metadata.distribution(self.main_package)
        package_path = self.resolve_package_path(dist)

        lock_path = (Path(package_path) / ".watchdog.lock").absolute()
        if lock_path.exists():
            raise WatchdogLockError(lock_path=lock_path)

        try:
            # Create the lock - this caller should now have exclusive access to the watchdog
            lock_path.touch()
            yield
        finally:
            # Remove the lock
            lock_path.unlink()

    def merge_paths(self, raw_paths: list[str]):
        """
        If one path is a subdirectory of another, we only want to watch the parent
        directory. This function merges the paths to avoid duplicate watchers.

        """
        paths = [Path(path).resolve() for path in raw_paths]

        # Parents should come before their subdirectories
        paths.sort(key=lambda path: len(path.parts))

        merged: list[Path] = []

        for path in paths:
            # Check if the current path is a subdirectory of any path in the merged list
            if not any(path.is_relative_to(parent) for parent in merged):
                merged.append(path)

        # Convert Path objects back to strings
        return [str(path) for path in merged]



from pathlib import Path

import pytest
from pydantic import BaseModel

from mountaineer.actions import passthrough
from mountaineer.app import AppController
from mountaineer.client_builder.openapi import OpenAPIDefinition
from mountaineer.controller import ControllerBase
from mountaineer.exceptions import APIException


def test_requires_render_return_value():
    """
    The AppController is in charge of validating our render return value. Since renders are not
    decorated, the best place to validate is during a mount.

    """

    class TestControllerWithoutRenderMarkup(ControllerBase):
        url = "/"
        view_path = "/page.tsx"

        def render(self):
            return None

    class TestControllerWithRenderMarkup(ControllerBase):
        url = "/"
        view_path = "/page.tsx"

        def render(self) -> None:
            return None

    app = AppController(view_root=Path(""))
    with pytest.raises(ValueError, match="must have a return type annotation"):
        app.register(TestControllerWithoutRenderMarkup())

    app.register(TestControllerWithRenderMarkup())


def test_generate_openapi():
    class ExampleSubModel(BaseModel):
        sub_value: str

    class ExampleException(APIException):
        status_code = 401
        invalid_reason: str
        sub_model: ExampleSubModel

    class ExampleController(ControllerBase):
        url = "/example"
        view_path = "/example.tsx"

        def render(self) -> None:
            pass

        @passthrough(exception_models=[ExampleException])
        def test_exception_action(self):
            pass

    app = AppController(view_root=Path(""))
    app.register(ExampleController())
    openapi_spec = app.generate_openapi()
    openapi_definition = OpenAPIDefinition(**openapi_spec)

    assert openapi_definition.components.schemas.keys() == {
        "TestExceptionActionResponse",
        "ExampleException",
        "ExampleSubModel",
    }


def test_update_ref_path():
    app = AppController(view_root=Path(""))

    # Test both dictionaries and lists
    fixed_schema = app._update_ref_path(
        {
            "components": {
                "schemas": {
                    "ExampleSubModel": {
                        "properties": {
                            "sub_value": {
                                "$ref": "#/defs/ExampleSubModel",
                            },
                            "list_values": {
                                "type": "array",
                                "items": [
                                    {
                                        "$ref": "#/defs/ExampleSubModel",
                                    }
                                ],
                            },
                        },
                    }
                }
            }
        }
    )

    assert isinstance(fixed_schema, dict)
    properties = fixed_schema["components"]["schemas"]["ExampleSubModel"]["properties"]
    assert properties["sub_value"]["$ref"] == "#/components/schemas/ExampleSubModel"
    assert (
        properties["list_values"]["items"][0]["$ref"]
        == "#/components/schemas/ExampleSubModel"
    )



import asyncio
from json import dumps as json_dumps
from logging import warning
from os import PathLike, environ
from pathlib import Path
from platform import machine as platform_machine
from platform import system as platform_system
from shutil import move as shutil_move
from subprocess import PIPE
from subprocess import run as subprocess_run
from tarfile import open as tarfile_open
from tempfile import TemporaryDirectory
from urllib.request import urlopen

from packaging import version
from tqdm import tqdm

from mountaineer.js_compiler.exceptions import BuildProcessException
from mountaineer.logging import LOGGER

ESBUILD_VERSION = "0.19.11"
URL_PATTERN = "https://registry.npmjs.org/@esbuild/{platform}/-/{filename}"
CACHE_PATH = Path("~/.cache/mountaineer/esbuild").expanduser()


class ESBuildWrapper:
    """
    Python shim for esbuild, only supports a subset of the spec.

    https://esbuild.github.io/api/#build

    """

    def __init__(self, allow_auto_download: bool = True):
        self.allow_auto_download = allow_auto_download

    async def bundle(
        self,
        *,
        entry_points: list[PathLike],
        outfile: PathLike,
        bundle: bool | None = None,
        loaders: dict[str, str] | None = None,
        output_format: str | None = None,
        global_name: str | None = None,
        sourcemap: bool | None = None,
        define: dict[str, str] | None = None,
        node_paths: list[str | Path] | None = None,
    ):
        # Make sure the output file path exists
        Path(outfile).parent.mkdir(parents=True, exist_ok=True)

        # Assemble and run a async subprocess with the above params
        command = [str(self.get_esbuild_path())]

        # We always expect to have entry points and outputs
        for entry_point in entry_points:
            command.append(str(entry_point))
        command.append(f"--outfile={outfile}")

        if bundle:
            command.append("--bundle")

        for ext, loader in (loaders or {}).items():
            command.append(f"--loader:{ext}={loader}")

        if output_format:
            command.append(f"--format={output_format}")

        if global_name:
            command.append(f"--global-name={global_name}")

        if sourcemap:
            command.append("--sourcemap")

        for key, value in (define or {}).items():
            command.append(f"--define:{key}={json_dumps(value)}")

        process = await asyncio.create_subprocess_exec(
            *command,
            stdout=PIPE,
            stderr=PIPE,
            env={
                **environ,
                "NODE_PATH": ":".join([str(path) for path in (node_paths or [])]),
            },
        )

        stdout, stderr = await process.communicate()

        if process.returncode != 0:
            raise BuildProcessException(f"esbuild error: {stderr.decode()}")

        return stdout.decode()

    def get_esbuild_path(self):
        """
        If esbuild isn't already available in the path, we download it based
        on the current platform.

        This follows the logic in their official install script:
        https://esbuild.github.io/dl/v0.19.11
        """

        # Determine if we already have esbuild installed
        installed_path = self.get_installed_path()
        if installed_path:
            return installed_path

        if not self.allow_auto_download:
            raise Exception(
                "esbuild is not installed and auto-download is disabled, bundle can't continue."
            )

        # Identify the platform
        system: str = platform_system()
        machine: str = platform_machine()

        # Map the system and machine to the esbuild platform
        if system == "Darwin":
            esbuild_platform = "darwin-arm64" if machine == "arm64" else "darwin-x64"
        elif system == "Linux":
            esbuild_platform = (
                "linux-arm64" if machine in ["arm64", "aarch64"] else "linux-x64"
            )
        elif system == "NetBSD":
            esbuild_platform = "netbsd-x64"
        elif system == "OpenBSD":
            esbuild_platform = "openbsd-x64"
        else:
            raise Exception(f"Unsupported platform: {system} {machine}")

        filename = f"{esbuild_platform}-{ESBUILD_VERSION}.tgz"
        url = URL_PATTERN.format(platform=esbuild_platform, filename=filename)

        with TemporaryDirectory() as temp_dir:
            temp_dir_path = Path(temp_dir)
            tgz_path = temp_dir_path / filename
            LOGGER.info(f"Downloading current version of esbuild from {url}")
            self.download_with_progress(url, tgz_path)

            # Extract the binary executable
            with tarfile_open(tgz_path, "r:gz") as tar:
                tar.extractall(path=temp_dir)

            # Make sure the cache path exists
            CACHE_PATH.parent.mkdir(parents=True, exist_ok=True)
            shutil_move(temp_dir_path / "package" / "bin" / "esbuild", CACHE_PATH)

        return CACHE_PATH

    def download_with_progress(self, url, filename):
        """
        Download a file with a progress bar.
        """
        with urlopen(url) as response, open(filename, "wb") as out_file:
            file_size = int(response.info().get("Content-Length"))
            chunk_size = 1024

            with tqdm(total=file_size, unit="B", unit_scale=True) as bar:
                for data in iter(lambda: response.read(chunk_size), b""):
                    out_file.write(data)
                    bar.update(len(data))

    def get_installed_path(self):
        """
        Check the local version of esbuild and issue a warning if it's outdated.i

        """
        # If they already have the cache path, we can assume they have the right version
        if Path(CACHE_PATH).exists():
            return CACHE_PATH

        result = subprocess_run(
            "esbuild --version", capture_output=True, text=True, shell=True
        )
        if result.returncode == 0 and version.parse(
            result.stdout.strip()
        ) != version.parse(ESBUILD_VERSION):
            warning(
                f"The locally found esbuild version is older than {ESBUILD_VERSION}, continuing anyway...",
                UserWarning,
            )

        # Get the path to this executable
        result = subprocess_run(
            "which esbuild", capture_output=True, text=True, shell=True
        )
        if result.returncode == 0:
            return result.stdout.strip()
        return False



import asyncio
import socket
from importlib import import_module
from multiprocessing import Event, Process, get_start_method, set_start_method
from multiprocessing.queues import Queue
from signal import SIGINT, signal
from threading import Thread
from time import sleep, time
from traceback import format_exception
from typing import Callable

from click import secho
from fastapi import Request
from pydantic.main import BaseModel

from mountaineer.app import AppController
from mountaineer.client_builder.builder import ClientBuilder
from mountaineer.controllers.exception_controller import ExceptionController
from mountaineer.logging import LOGGER
from mountaineer.watch import CallbackDefinition, CallbackType, PackageWatchdog
from mountaineer.watch_server import WatcherWebservice
from mountaineer.webservice import UvicornThread


class IsolatedRunserverConfig(BaseModel):
    entrypoint: str
    port: int
    live_reload_port: int


class IsolatedWatchConfig(BaseModel):
    webcontroller: str


class IsolatedEnvProcess(Process):
    """
    We need a fully separate process for our runserver and watch, so we're able to re-import
    all of the dependent files when there are changes.

    """

    def __init__(
        self,
        watch_config: IsolatedWatchConfig,
        runserver_config: IsolatedRunserverConfig | None = None,
        build_notification_channel: Queue | None = None,
    ):
        super().__init__()

        if runserver_config is None and watch_config is None:
            raise ValueError("Must provide either runserver_config or watch_config")

        self.runserver_config = runserver_config
        self.watch_config = watch_config
        self.close_signal = Event()
        self.build_notification_channel = build_notification_channel

    def run(self):
        app_controller = import_from_string(self.watch_config.webcontroller)
        if not isinstance(app_controller, AppController):
            raise ValueError(
                f"Expected {self.watch_config.webcontroller} to be an instance of AppController"
            )

        # Mount our exceptions controller, since we'll need these artifacts built
        # as part of the JS build phase
        self.exception_controller = ExceptionController()
        app_controller.register(self.exception_controller)
        app_controller.app.exception_handler(Exception)(self.handle_dev_exception)

        # Finish the build before we start the server since the server launch is going to sniff
        # for the built artifacts
        if self.watch_config is not None:
            secho("Starting build...", fg="yellow")
            start = time()

            js_compiler = ClientBuilder(
                app_controller,
                live_reload_port=(
                    self.runserver_config.live_reload_port
                    if self.runserver_config
                    else None
                ),
            )
            js_compiler.build()
            secho(f"Build finished in {time() - start:.2f} seconds", fg="green")

            self.alert_notification_channel()

        if self.runserver_config is not None:
            thread = UvicornThread(
                app=app_controller.app,
                port=self.runserver_config.port,
            )
            thread.start()
            try:
                self.close_signal.wait()
            except KeyboardInterrupt:
                pass
            thread.stop()
            thread.join()

        LOGGER.debug("IsolatedEnvProcess finished")

    def alert_notification_channel(self):
        """
        Alerts the notification channel of a build update, once the server
        comes back online. Before this the client might refresh and get a blank
        page because the server hasn't yet booted.
        """

        def wait_for_server():
            # No need to do anything if we don't have a notification channel
            if self.runserver_config is None:
                return
            if self.build_notification_channel is None:
                return

            # Loop until there is something bound to the runserver_config.port
            start = time()
            LOGGER.debug("Waiting for server to come online")
            while True:
                try:
                    with socket.create_connection(
                        ("localhost", self.runserver_config.port)
                    ):
                        break
                except ConnectionRefusedError:
                    sleep(0.1)
            LOGGER.debug(f"Server took {time() - start:.2f} seconds to come online")

            # Buffer to make sure the server is fully booted
            sleep(0.5)

            if self.build_notification_channel:
                self.build_notification_channel.put(True)

        alert_thread = Thread(target=wait_for_server)
        alert_thread.start()

    def stop(self, hard_timeout: float = 5.0):
        """
        Client-side stop method to shut down the running process.
        """
        # If we've already stopped, don't try to stop again
        if not self.is_alive():
            return
        if self.runserver_config is not None:
            self.close_signal.set()
            # Try to give the server time to shut down gracefully
            while self.is_alive() and hard_timeout > 0:
                self.join(1)
                hard_timeout -= 1

            if hard_timeout == 0:
                secho(
                    f"Server shutdown reached hard timeout deadline: {self.is_alive()}",
                    fg="red",
                )

        # As a last resort we send a hard termination signal
        if self.is_alive():
            self.terminate()

    async def handle_dev_exception(self, request: Request, exc: Exception):
        return await self.exception_controller._generate_html(
            global_metadata=None,
            exception=str(exc),
            stack="".join(format_exception(exc)),
        )


def handle_watch(
    *,
    package: str,
    webcontroller: str,
    subscribe_to_fizl: bool = False,
):
    """
    Watch the file directory and rebuild auto-generated files.

    :param client_package: "ci_webapp"
    :param client_controller: "ci_webapp.app:controller"

    """
    update_multiprocessing_settings()

    current_process: IsolatedEnvProcess | None = None

    def update_build():
        nonlocal current_process

        if current_process is not None:
            # Stop the current process if it's running
            current_process.stop()

        current_process = IsolatedEnvProcess(
            watch_config=IsolatedWatchConfig(webcontroller=webcontroller),
        )
        current_process.start()

    watchdog = build_common_watchdog(
        package, update_build, subscribe_to_fizl=subscribe_to_fizl
    )
    watchdog.start_watching()


def handle_runserver(
    *,
    package: str,
    webservice: str,
    webcontroller: str,
    port: int,
    subscribe_to_fizl: bool = False,
):
    """
    :param client_package: "ci_webapp"
    :param client_webservice: "ci_webapp.app:app"
    :param client_controller: "ci_webapp.app:controller"

    """
    update_multiprocessing_settings()

    current_process: IsolatedEnvProcess | None = None

    # Start the webservice - it should persist for the lifetime of the
    # runserver, so a single websocket frontend can be notified across
    # multiple builds
    watcher_webservice = WatcherWebservice()
    watcher_webservice.start()

    def update_webservice():
        asyncio.run(watcher_webservice.broadcast_listeners())

    def update_build():
        nonlocal current_process

        if current_process is not None:
            # Stop the current process if it's running
            current_process.stop()

        current_process = IsolatedEnvProcess(
            runserver_config=IsolatedRunserverConfig(
                entrypoint=webservice,
                port=port,
                live_reload_port=watcher_webservice.port,
            ),
            watch_config=IsolatedWatchConfig(webcontroller=webcontroller),
            build_notification_channel=watcher_webservice.notification_queue,
        )
        current_process.start()

    # Install a signal handler to catch SIGINT and try to
    # shut down gracefully
    def graceful_shutdown(signum, frame):
        if current_process is not None:
            current_process.stop()
        if watcher_webservice is not None:
            watcher_webservice.stop()
        secho("Services shutdown, now exiting...", fg="yellow")
        exit(0)

    signal(SIGINT, graceful_shutdown)

    watchdog = build_common_watchdog(
        package, update_build, subscribe_to_fizl=subscribe_to_fizl
    )
    watchdog.start_watching()


def update_multiprocessing_settings():
    """
    fork() is still the default on Linux, and can result in stalls with our asyncio
    event loops. For consistency and expected behavior, try to switch to spawn()
    if it's not already enabled.

    """
    if (spawn_method := get_start_method()) and spawn_method != "spawn":
        LOGGER.warning(
            f"The watch command should be run with the spawn start method, but it's currently set to {spawn_method}",
        )

        try:
            set_start_method("spawn", force=True)
            LOGGER.info("Start method is now set to 'spawn'.")
        except RuntimeError as e:
            # We might catch errors if set_start_method('spawn') is called
            # in an inappropriate context or after multiprocessing has started, which
            # is not allowed.
            LOGGER.error(f"Cannot change the start method after it has been used: {e}")


def import_from_string(import_string: str):
    """
    Given a string to the package (like "ci_webapp.app:controller") import the
    actual variable
    """
    module_name, attribute_name = import_string.split(":")
    module = import_module(module_name)
    return getattr(module, attribute_name)


def build_common_watchdog(
    client_package: str,
    callback: Callable,
    subscribe_to_fizl: bool,
):
    """
    Useful creation class to build a watchdog the common client class
    and our internal package.

    :param subscribe_to_fizl: If True, we'll also subscribe to the mountaineer package
    changes in the local environment. This is helpful for local development of the core
    package concurrent with a downstream client application.

    """
    return PackageWatchdog(
        client_package,
        dependent_packages=["mountaineer", "mountaineer_auth", "mountaineer_daemons"]
        if subscribe_to_fizl
        else [],
        callbacks=[
            CallbackDefinition(
                CallbackType.CREATED | CallbackType.MODIFIED,
                callback,
            )
        ],
        # We want to generate a build on the first load
        run_on_bootup=True,
    )



from warnings import filterwarnings

import pytest


@pytest.fixture(autouse=True)
def ignore_httpx_deprecation_warnings():
    # Ignore httpx deprecation warnings until fastapi updates its internal test constructor
    # https://github.com/encode/httpx/blame/master/httpx/_client.py#L678
    filterwarnings("ignore", category=DeprecationWarning, module="httpx.*")



from abc import ABC, abstractmethod
from typing import Any, Coroutine

from pydantic import BaseModel

from mountaineer.controller import ControllerBase
from mountaineer.paths import ManagedViewPath


class ClientBundleMetadata(BaseModel):
    live_reload_port: int | None = None


class ClientBuilderBase(ABC):
    """
    Base class for client builders. When mounted to an AppController, these build plugins
    will be called for every file defined in the view/app directory. It's up to the plugin
    whether to handle the incoming file.

    """

    @abstractmethod
    def handle_file(
        self,
        current_path: ManagedViewPath,
        controller: ControllerBase | None,
        metadata: ClientBundleMetadata,
    ) -> None | Coroutine[Any, Any, None]:
        """
        Only direct controller views are called with (view, controller) inputs. Otherwise we do a
        recursive search of the raw files on disk with controller=None.

        """
        pass



from contextlib import contextmanager

from pydantic._internal._model_construction import ModelMetaclass
from pydantic_settings import BaseSettings


class ConfigMeta(ModelMetaclass):
    def __call__(cls, *args, **kwargs):
        instance = super().__call__(*args, **kwargs)
        register_config(instance)
        return instance


class ConfigBase(BaseSettings, metaclass=ConfigMeta):
    model_config = {"frozen": True}


# One global config object
APP_CONFIG: ConfigBase | None = None


def register_config(config: ConfigBase):
    global APP_CONFIG

    if APP_CONFIG is not None:
        raise ValueError("Config already registered")

    APP_CONFIG = config


def unregister_config():
    global APP_CONFIG
    APP_CONFIG = None


def get_config() -> ConfigBase:
    if APP_CONFIG is None:
        raise ValueError(
            "Configuration not registered. Either:\n"
            "1. Call register_config() with your BaseSettings class\n"
            "2. Make sure your BaseSettings is imported so the ConfigMeta can auto-register"
        )

    return APP_CONFIG


@contextmanager
def register_config_in_context(config: ConfigBase):
    """
    Change the global config object to the given config object
    temporarily. Useful for unit testing.
    """
    global APP_CONFIG
    previous_config = APP_CONFIG
    APP_CONFIG = config
    try:
        yield
    finally:
        APP_CONFIG = previous_config



from pathlib import Path

from jinja2 import Template
from pydantic import BaseModel

from create_mountaineer_app.templates import get_template_path


class ProjectMetadata(BaseModel):
    project_name: str
    author_name: str
    author_email: str
    use_poetry: bool
    use_tailwind: bool
    project_path: Path

    # If specified, will install mountaineer in development mode pointing to a local path
    # This is useful for testing changes to mountaineer itself
    mountaineer_dev_path: Path | None = None


class TemplateOutput(BaseModel):
    content: str
    path: str


def format_template(name: str, project_metadata: ProjectMetadata) -> TemplateOutput:
    """
    Takes in a template path (relative to /templates) and returns the formatted
    template contents and the final path of the file.

    We support:
        - Jinja templating within the file
        - Bracket syntax in filenames, like /path/to/[project_name]/file.txt

    """
    path = get_template_path("project") / name
    if not path.exists():
        raise FileNotFoundError(f"Template file {path} does not exist")

    metadata_variables = project_metadata.model_dump()

    template = Template(path.read_text())
    content = template.render(metadata_variables)

    output_name = name
    for key, value in metadata_variables.items():
        output_name = output_name.replace(f"[{key}]", str(value))

    return TemplateOutput(content=content, path=output_name)



from pathlib import Path

from create_mountaineer_app.generation import ProjectMetadata, format_template


def test_path_url_replacement():
    metadata = ProjectMetadata(
        project_name="TEST_PROJECT_NAME",
        author_name="TEST_AUTHOR",
        author_email="TEST_EMAIL",
        use_tailwind=True,
        use_poetry=True,
        project_path=Path("fake-path"),
        mountaineer_dev_path=None,
    )
    bundle = format_template("[project_name]/app.py", metadata)
    assert bundle.path == "TEST_PROJECT_NAME/app.py"



from pathlib import Path
from re import match as re_match

import questionary
from click import command, option, secho

from create_mountaineer_app.builder import build_project
from create_mountaineer_app.environments.poetry import PoetryEnvironment
from create_mountaineer_app.external import (
    get_git_user_info,
)
from create_mountaineer_app.generation import ProjectMetadata


def prompt_should_use_poetry():
    input_use_poetry = questionary.confirm(
        "Use poetry for dependency management? [Yes]", default=True
    ).unsafe_ask()

    if not input_use_poetry:
        return input_use_poetry

    # Helpful utilities to wrap poetry logic and lifecycle
    poetry_environment = PoetryEnvironment()

    # Determine if the user already has poetry
    if not poetry_environment.has_provider():
        input_install_poetry = questionary.confirm(
            "Poetry is not installed. Install poetry?", default=True
        ).unsafe_ask()
        if input_install_poetry:
            secho("Installing poetry...")
            try:
                poetry_environment.install_provider()
            except Exception as e:
                secho(f"Error installing poetry: {e}", fg="red")
                raise e

    return input_use_poetry


def prompt_author() -> tuple[str, str]:
    while True:
        git_name, git_email = get_git_user_info()
        default_author = (
            f"{git_name} <{git_email}>" if git_name and git_email else "Name <email>"
        )
        author_combined = (
            questionary.text(f"Author [{default_author}]").unsafe_ask()
            or default_author
        )

        # Parse the author into a name and email
        matched_obj = re_match(r"(.*) <(.*)>", author_combined)

        if not matched_obj:
            secho(
                f"Author must be in the format 'Name <email>', not '{author_combined}'",
                fg="red",
            )
            continue

        return matched_obj.group(1), matched_obj.group(2)


@command()
@option("--output-path", help="The output path for the bundled files.")
@option("--mountaineer-dev-path", help="The path to the Mountaineer dev environment.")
def main(output_path: str | None, mountaineer_dev_path: str | None):
    """
    Create a new Mountaineer project.

    Most of the configuration is handled interactively. We have a few CLI arguments for rarer options that can configure the installation behavior.

    :param output_path: The output path for the bundled files, if not the current directory.

    """
    input_project_name = (
        questionary.text("Project name [my-project]:").unsafe_ask() or "my-project"
    )
    input_author_name, input_author_email = prompt_author()
    input_use_poetry = prompt_should_use_poetry()
    input_use_tailwind = questionary.confirm(
        "Use Tailwind CSS? [Yes]", default=True
    ).unsafe_ask()

    secho("\nCreating project...", fg="green")

    project_path = Path(output_path) if output_path else Path.cwd() / input_project_name
    project_path = project_path.resolve()

    metadata = ProjectMetadata(
        project_name=input_project_name.replace(" ", "_").replace("-", "_"),
        author_name=input_author_name,
        author_email=input_author_email,
        use_poetry=input_use_poetry,
        use_tailwind=input_use_tailwind,
        project_path=project_path,
        mountaineer_dev_path=Path(mountaineer_dev_path).resolve()
        if mountaineer_dev_path
        else None,
    )

    build_project(metadata)


if __name__ == "__main__":
    main()



from typing import Any, Type, cast, get_type_hints

from fastapi import HTTPException
from pydantic import BaseModel, Field, create_model

# Keys that are used to initialize the HTTPException model
HTTPExceptionKeys = ["status_code", "detail", "headers"]


class APIExceptionInternalModelBase(BaseModel):
    """
    Superclass used for our synthetic internal errors. This class sets
    the required parameters and default model configuration for use in
    validating user inputs to APIException(**kwargs)

    """

    status_code: int
    detail: str
    headers: dict[str, str]

    model_config = {"extra": "forbid"}


class InternalModelMeta(type):
    def __new__(mcs, name, bases, namespace):
        cls = super().__new__(mcs, name, bases, namespace)
        cls._create_internal_model()
        return cls

    def _create_internal_model(cls):
        type_hints = get_type_hints(cls)

        fields = {
            key: (
                key_type,
                getattr(cls, key, Field(default_factory=lambda: None)),
            )
            for key, key_type in type_hints.items()
            if key not in ["InternalModel", "internal_model"]
        }
        cls.InternalModel = create_model(
            # Mirror the class name so our OpenAPI objects are as the user specifies
            # for their exception class
            cls.__name__,
            __base__=APIExceptionInternalModelBase,
            **cast(Any, fields),
        )

    def __call__(cls, *args, **kwargs):
        # Override the __call__ method to instantiate models like Pydantic does
        # Use the internal model for validation and instantiation
        internal_model = cls.InternalModel(**kwargs)
        instance = super().__call__(
            **{
                key: value
                for key, value in internal_model.model_dump().items()
                if key in HTTPExceptionKeys
            }
        )
        setattr(instance, "internal_model", internal_model)
        return instance


class APIException(HTTPException, metaclass=InternalModelMeta):
    status_code: int = 500
    detail: str = "A server error occurred"
    headers: dict[str, str] = Field(default_factory=dict)

    InternalModel: Type[APIExceptionInternalModelBase]
    internal_model: APIExceptionInternalModelBase

    # We can't synthetically create an API contract with the instance variables
    # (pydantic uses a plugin to get their IDE typehints)
    # Any issues with the input constructor will raise a runtime error
    # versus being statically checked
    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)



from pathlib import Path

from click import secho

from create_mountaineer_app.environments.base import EnvironmentBase
from create_mountaineer_app.environments.poetry import PoetryEnvironment
from create_mountaineer_app.environments.venv import VEnvEnvironment
from create_mountaineer_app.external import (
    has_npm,
    npm_install,
)
from create_mountaineer_app.generation import ProjectMetadata, format_template
from create_mountaineer_app.templates import get_template_path

IGNORE_FILES = {"__pycache__", "node_modules"}
ALLOW_HIDDEN_FILES = {".env"}


def environment_from_metadata(metadata: ProjectMetadata) -> EnvironmentBase:
    if metadata.use_poetry:
        return PoetryEnvironment()
    else:
        return VEnvEnvironment()


def should_copy_path(path: Path):
    """
    Determine whether we should copy the template path to our final project.
    We need to ignore certain build-time directories that don't actually
    have code logic.

    - Ignore explicitly ignored files
    - Ignore hidden files and folders

    """
    for part in path.parts:
        if part in IGNORE_FILES:
            return False
        if part.startswith(".") and part not in ALLOW_HIDDEN_FILES:
            return False
    return True


def build_project(metadata: ProjectMetadata):
    template_base = get_template_path("project")

    for template_path in template_base.glob("**/*"):
        if template_path.is_dir() or not should_copy_path(template_path):
            continue

        # Internally, format_template will re-look up the template. Convert the full glob path into a relative template path.
        template_name = str(template_path.relative_to(template_base))
        try:
            output_bundle = format_template(template_name, metadata)
        except Exception as e:
            secho(f"Error formatting {template_path}: {e}", fg="red")
            raise e

        if not output_bundle.content.strip():
            secho(
                f"No content detected in {output_bundle.path}, skipping...", fg="yellow"
            )
            continue

        secho(f"Creating {output_bundle.path}")
        full_output = metadata.project_path / output_bundle.path
        full_output.parent.mkdir(parents=True, exist_ok=True)
        full_output.write_text(output_bundle.content)

    secho(f"Project created at {metadata.project_path}", fg="green")

    environment = environment_from_metadata(metadata)

    try:
        environment.install_project(metadata.project_path)
    except Exception as e:
        secho(f"Error installing python dependencies: {e}", fg="red")

    if has_npm():
        npm_install(metadata.project_path / metadata.project_name / "views")
    else:
        secho(
            "npm is not installed and is required to install React dependencies.",
            fg="red",
        )
